#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import u
u.default_format = '~P'
import numpy as np
from pint.testing import assert_allclose


# tests


def test_structure_str(structure):
    structure.__str__()


# test_visualize() fails on certain virtual machine due to TCL error


def test_structure_get_hash(structure):
    structure.get_hash()


def test_get_number_of_sub_structures(structure):
    assert structure.get_number_of_sub_structures() == 5


def test_get_number_of_layers(structure):
    assert structure.get_number_of_layers() == 280


def test_get_number_of_unique_layers(structure):
    assert structure.get_number_of_unique_layers() == 4


def test_get_thickness(structure):
    assert_allclose(structure.get_thickness(), 150*u.nm, rtol=1e-2)


def test_get_unique_layers(structure):
    layer_ids, _ = structure.get_unique_layers()
    assert layer_ids == ['amorphous_layer_Fe', 'amorphous_layer_O', 'unit_cell_Fe', 'unit_cell_O']


def test_get_layer_vectors(structure):
    structure.get_layer_vectors()


def test_get_all_positions_per_unique_layer(structure):
    pos = structure.get_all_positions_per_unique_layer()
    assert np.allclose(pos['amorphous_layer_Fe'], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])


def test_get_distances_of_layers(structure):
    d_start, _, _ = structure.get_distances_of_layers()
    assert_allclose(np.sum(d_start), 2.22e-05*u.m, rtol=1e-2)


def test_get_distances_of_interfaces(structure):
    assert_allclose(np.sum(structure.get_distances_of_interfaces()), 3.68e-6*u.m, rtol=1e-2)


def test_interp_distance_at_interfaces(structure):
    structure.interp_distance_at_interfaces(N=11)


def test_get_layer_property_vector(structure):
    structure.get_layer_property_vector('density')
    structure.get_layer_property_vector('_density')
    structure.get_layer_property_vector('heat_capacity')
    structure.get_layer_property_vector('spring_const')


def test_get_layer_handle(structure):
    structure.get_layer_handle(0)


def test_reverse(structure):
    structure.reverse()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import XrayKin, XrayDyn, XrayDynMag
from udkm1Dsim import u
import numpy as np
import pytest
from pint.testing import assert_allclose


# fixtures


@pytest.fixture(scope='module')
def xray_kin(structure_crystalline, tmp_path_factory):
    return XrayKin(structure_crystalline, force_recalc=True,
                   cache_dir=tmp_path_factory.mktemp('cache'),
                   save_data=True, disp_messages=True, progress_bar=True,
                   )


@pytest.fixture(scope='module')
def xray_dyn(structure_crystalline, tmp_path_factory):
    return XrayDyn(structure_crystalline, force_recalc=True,
                   cache_dir=tmp_path_factory.mktemp('cache'),
                   save_data=True, disp_messages=True, progress_bar=True,
                   )


@pytest.fixture(scope='module')
def xray_dyn_mag(structure, tmp_path_factory):
    return XrayDynMag(structure, force_recalc=True,
                      cache_dir=tmp_path_factory.mktemp('cache'),
                      save_data=True, disp_messages=True, progress_bar=True,
                      )


@pytest.fixture(scope='module')
def delays():
    return np.r_[-1:10:1]*u.ps


@pytest.fixture(scope='module')
def distances(structure_amorph):
    dists, _, _ = structure_amorph.get_distances_of_layers()
    return dists


@pytest.fixture(scope='module')
def distances_mixed(structure):
    dists, _, _ = structure.get_distances_of_layers()
    return dists


@pytest.fixture(scope='module')
def temp_map(delays, distances):
    return 50*np.ones([len(delays), len(distances)])


@pytest.fixture(scope='module')
def strain_map(delays, distances):
    return 0.01*np.ones([len(delays), len(distances)])


@pytest.fixture(scope='module')
def strain_map_mixed(delays, distances_mixed):
    return 0.01*np.ones([len(delays), len(distances_mixed)])


@pytest.fixture(scope='module')
def magnetization_map(delays, distances_mixed):
    return np.zeros([len(delays), len(distances_mixed), 3])


# tests


# XrayKin


def text_xray_kin_str(xray_kin):
    xray_kin.__str__()


def test_xray_kin_update_experiment(xray_kin):
    # energy, wl, or k must be set first
    with pytest.raises(IndexError):
        xray_kin.theta = 45*u.deg
    xray_kin.energy = 1000*u.eV
    xray_kin.theta = 45*u.deg
    assert_allclose(xray_kin.qz, np.array([[7.1669]])/u.nm, rtol=1e-2)
    xray_kin.qz = 1/u.nm
    assert_allclose(xray_kin.theta, np.array([[5.66]])*u.deg, rtol=1e-2)
    assert_allclose(xray_kin.wl, np.array([1.24])*u.nm, rtol=1e-2)
    assert_allclose(xray_kin.k, np.array([5.07])/u.nm, rtol=1e-2)
    xray_kin.k = 1/u.nm
    xray_kin.wl = 1*u.nm


def text_xray_kin_set_polarization(xray_kin):
    xray_kin.set_polarization(0, 0)
    xray_kin.set_polarization(1, 1)
    xray_kin.set_polarization(2, 0)
    xray_kin.set_polarization(3, 0)
    xray_kin.set_polarization(4, 0)


def test_xray_kin_get_polarization_factor(xray_kin):
    xray_kin.get_polarization_factor(0)


def test_xray_kin_get_hash(xray_kin, strain_map):
    strain_vectors = {}
    xray_kin.get_hash(strain_vectors, strain_map=strain_map)


def test_xray_kin_get_uc_structure_factor(xray_kin, unit_cell_iron):
    assert_allclose(xray_kin.get_uc_structure_factor(1000, 1, unit_cell_iron, strain=0.01),
                    -15.52-11.72j, rtol=1e-2)


def test_xray_kin_homogeneous_reflectivity(xray_kin):
    xray_kin.homogeneous_reflectivity()


# XrayDyn


def text_xray_dyn_str(xray_dyn):
    xray_kin.__str__()


def text_xray_dyn_set_polarization(xray_dyn):
    xray_dyn.set_polarization(0, 0)
    xray_dyn.set_polarization(1, 1)
    xray_dyn.set_polarization(2, 0)
    xray_dyn.set_polarization(3, 0)
    xray_dyn.set_polarization(4, 0)


def test_xray_dyn_get_hash(xray_dyn, strain_map, temp_map):
    strain_vectors = {}
    xray_dyn.get_hash(strain_vectors, strain_map=strain_map, temp_map=temp_map)


def test_xray_dyn_homogeneous_reflectivity(xray_dyn):
    xray_dyn.energy = 8000*u.eV
    xray_dyn.theta = np.r_[1:10]*u.deg
    xray_dyn.homogeneous_reflectivity()
    xray_dyn.homogeneous_reflectivity(strains=np.zeros([2]))
    xray_dyn.homogeneous_reflectivity(temps=300*np.ones([2]))
    xray_dyn.homogeneous_reflectivity(strains=np.zeros([2]), temps=300*np.ones([2]))
    with pytest.raises(IndexError):
        xray_dyn.homogeneous_reflectivity(strains=np.zeros([3]))
    with pytest.raises(IndexError):
        xray_dyn.homogeneous_reflectivity(temps=300*np.ones([3]))
    with pytest.raises(IndexError):
        xray_dyn.homogeneous_reflectivity(temps=300*np.ones([2, 2]))


def test_xray_dyn_inhomogeneous_reflectivity(xray_dyn, strain_map, temp_map):
    xray_dyn.energy = 8000*u.eV
    xray_dyn.theta = np.r_[1:10]*u.deg
    xray_dyn.inhomogeneous_reflectivity(strain_map)
    xray_dyn.inhomogeneous_reflectivity(strain_map, temp_map=temp_map)
    strain_vectors = [np.r_[0:1], np.r_[0:1]]
    with pytest.warns(UserWarning):
        xray_dyn.inhomogeneous_reflectivity(strain_map, strain_vectors=strain_vectors,
                                            temp_map=temp_map)
    xray_dyn.inhomogeneous_reflectivity(strain_map, strain_vectors=strain_vectors)
    xray_dyn.force_recalc = False
    xray_dyn.inhomogeneous_reflectivity(strain_map, strain_vectors=strain_vectors)
    xray_dyn.force_recalc = True
    with pytest.raises(ValueError):
        xray_dyn.inhomogeneous_reflectivity(strain_map, strain_vectors=strain_vectors,
                                            calc_type='parallel')
    with pytest.raises(NotImplementedError):
        xray_dyn.inhomogeneous_reflectivity(strain_map, strain_vectors=strain_vectors,
                                            calc_type='distributed')
    with pytest.raises(TypeError):
        xray_dyn.inhomogeneous_reflectivity([])
    with pytest.raises(TypeError):
        xray_dyn.inhomogeneous_reflectivity(strain_map, temp_map=[])
    with pytest.raises(TypeError):
        xray_dyn.inhomogeneous_reflectivity(strain_map, strain_vectors=np.array([]))
    with pytest.raises(ValueError):
        xray_dyn.inhomogeneous_reflectivity(strain_map, temp_map=np.ones([11, 20, 2]))


# XrayDynMag


def text_xray_dyn_mag_str(xray_dyn_mag):
    xray_dyn_mag.__str__()


def text_xray_dyn_mag_set_polarization(xray_dyn_mag):
    xray_dyn_mag.set_polarization(0, 0)
    xray_dyn_mag.set_polarization(1, 1)
    xray_dyn_mag.set_polarization(2, 0)
    xray_dyn_mag.set_polarization(3, 0)
    xray_dyn_mag.set_polarization(4, 0)


def test_xray_dyn_mag_get_hash(xray_dyn_mag, strain_map_mixed, magnetization_map):
    xray_dyn_mag.get_hash(strain_map=strain_map_mixed, magnetization_map=magnetization_map)


def test_xray_dyn_mag_homogeneous_reflectivity(xray_dyn_mag):
    xray_dyn_mag.energy = 800*u.eV
    xray_dyn_mag.theta = np.r_[1:10]*u.deg
    xray_dyn_mag.homogeneous_reflectivity()


def test_xray_dyn_mag_inhomogeneous_reflectivity(xray_dyn_mag, strain_map_mixed,
                                                 magnetization_map):
    xray_dyn_mag.energy = 800*u.eV
    xray_dyn_mag.theta = np.r_[1:10]*u.deg
    xray_dyn_mag.inhomogeneous_reflectivity(strain_map=strain_map_mixed,
                                            magnetization_map=magnetization_map)
    xray_dyn_mag.force_recalc = False
    xray_dyn_mag.inhomogeneous_reflectivity(strain_map=strain_map_mixed,
                                            magnetization_map=magnetization_map)
    xray_dyn_mag.force_recalc = True
    with pytest.raises(ValueError):
        xray_dyn_mag.inhomogeneous_reflectivity(strain_map=strain_map_mixed,
                                                magnetization_map=magnetization_map,
                                                calc_type='parallel')
    with pytest.raises(NotImplementedError):
        xray_dyn_mag.inhomogeneous_reflectivity(strain_map=strain_map_mixed,
                                                magnetization_map=magnetization_map,
                                                calc_type='distributed')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = ['Structure']

__docformat__ = 'restructuredtext'

from .layers import AmorphousLayer, UnitCell
from .. import u, Q_
from ..helpers import make_hash_md5, finderb
import itertools
import numpy as np


class Structure:
    """Structure

    Structure representation which holds various sub_structures.

    Each sub_structure can be either a layer of :math:`N` UnitCell or
    AmorphousLayer instances or a structure by itself.
    It is possible to recursively build up 1D structures.

    Args:
        name (str): name of the sample.

    Attributes:
        name (str): name of sample.
        sub_structures (list[AmorphousLayer, UnitCell, Structure]): list of
            structures in sample.
        substrate (Structure): structure of the substrate.
        num_sub_systems (int): number of subsystems for heat and phonons
           (electronic, lattice, spins, ...).

    """

    def __init__(self, name):
        self.name = name
        self.num_sub_systems = 1
        self.sub_structures = []
        self.substrate = []
        self.roughness = 0.0*u.nm

    def __str__(self, tabs=0):
        """String representation of this class"""
        tab_str = tabs*'\t'

        class_str = tab_str + 'Structure properties:\n\n'
        class_str += tab_str + 'Name   : {:s}\n'.format(self.name)

        if len(self.sub_structures) == 0:
            class_str += tab_str + 'Structure is empty\n----\n'
            return class_str

        class_str += tab_str + 'Thickness : {:.4g~P}\n'.format(self.get_thickness().to('nm'))
        class_str += tab_str + 'Roughness : {:.4g~P}\n'.format(self.roughness.to('nm'))
        class_str += tab_str + '----\n'
        # traverse all substructures
        for sub_structure in self.sub_structures:
            if isinstance(sub_structure[0], (AmorphousLayer, UnitCell)):
                # the substructure is an unitCell
                class_str += tab_str + '{:d} times {:s}: {:.4g~P}\n'.format(
                        sub_structure[1],
                        sub_structure[0].name,
                        sub_structure[1]*sub_structure[0].thickness.to('nm'))
            else:
                # the substructure is a structure instance by itself
                # call the display() method recursively
                class_str += tab_str + 'sub-structure {:d} times:\n'.format(
                       sub_structure[1])
                class_str += sub_structure[0].__str__(tabs+1)
        class_str += tab_str + '----\n'
        # check for a substrate
        if isinstance(self.substrate, Structure):
            class_str += tab_str + 'Substrate:\n'
            class_str += tab_str + '----\n'
            class_str += tab_str + '{:d} times {:s}: {:.4g~P}\n'.format(
                    self.substrate.sub_structures[0][1],
                    self.substrate.sub_structures[0][0].name,
                    self.substrate.sub_structures[0][1]
                    * self.substrate.sub_structures[0][0].thickness.to('nm'))
        else:
            class_str += tab_str + 'no substrate\n'
        return class_str

    def visualize(self, block=True, unit='nm', fig_size=[20, 1], cmap='Set1', linewidth=0.1,
                  show=True):
        """visualize

        Simple visualization of the structure.

        Args:
            unit (str): SI unit of the distance of the Structure. Defaults to
                'nm'.
            fig_size (list[float]): figure size of the visualization plot.
                Defaults to [20, 1].
            cmap (str): Matplotlib colormap for colors of layers.
            linewidth (float): line width of the patches.
            show (boolean): show visualization plot at the end.

        """
        import matplotlib.pyplot as plt
        from matplotlib import patches
        from matplotlib import cm

        _, d_end, _ = self.get_distances_of_layers(True)  # distance vector of all layers
        layer_interfaces = np.append(0, d_end.to(unit).magnitude)  # Append zero at the start
        thickness = np.max(layer_interfaces)

        layer_ids = self.get_unique_layers()[0]
        N = len(layer_ids)  # number of unique layers

        colortable = {}
        for i in range(N):
            colortable[layer_ids[i]] = cm.get_cmap(cmap)(i)

        plt.figure(figsize=fig_size)
        ax = plt.axes()

        for i, name in enumerate(self.get_layer_vectors()[1]):
            col = colortable.get(name, 'k')
            rect = patches.Rectangle((layer_interfaces[i], 0), np.diff(layer_interfaces)[i], 1,
                                     linewidth=linewidth, facecolor=col, edgecolor='k')
            ax.add_patch(rect)

        plt.xlim(0, thickness)
        plt.ylim(0, 1)
        plt.xlabel('Distance ({:s})'.format(unit))
        plt.yticks([], [])

        # add labels for legend
        for layer_id, col in colortable.items():
            plt.plot(0, 0, color=col, label=layer_id)

        leg = plt.legend(bbox_to_anchor=(0., 1.08, 1, .102), frameon=False, ncol=8)

        for line in leg.get_lines():
            line.set_linewidth(8.0)

        if show:
            plt.show(block=block)

    def get_hash(self, **kwargs):
        """get_hash

        Create an unique hash from all layer IDs in the correct order in the
        structure as well as the corresponding material properties which are
        given by the `kwargs`.

        `types='all'` is problematic, as function handles will be include,
        which will always change on recreation. Following errors from the
        GitHub test-suite, hashes do not seem to match across different OS
        and/or python versions.

        Args:
            **kwargs (list[str]): types of requested properties..

        Returns:
            hash (str): unique hash.

        """
        param = []
        layers = self.get_unique_layers()
        for layer in layers[1]:
            param.append(layer.get_property_dict(**kwargs))

        _, IDs, _ = self.get_layer_vectors()
        param.append(IDs)
        return make_hash_md5(param)

    def add_sub_structure(self, sub_structure, N=1):
        """add_sub_structure

        Add a sub_structure of :math:`N` layers or sub-structures to the
        structure.

        Args:
            sub_structure (AmorphousLayer, UnitCell, Structure):
               amorphous layer, unit cell, or structure to add as sub-structure.
            N (int): number or repetitions.

        """
        # check of the sub_structure is an instance of the unitCell or
        # structure class
        if not isinstance(sub_structure, (AmorphousLayer, UnitCell, Structure)):
            raise ValueError('Class '
                             + type(sub_structure).__name__
                             + ' is no possible sub structure. '
                             + 'Only AmorphousLayer, UnitCell, and '
                             + 'Structure classes are allowed!')

        # if a structure is added as a sub_structure, the sub_structure
        # can not have a substrate
        if isinstance(sub_structure, Structure):
            if sub_structure.substrate:
                raise ValueError('No substrate in sub_structure allowed!')

        # check the number of subsystems of the sub_structure
        if ((self.num_sub_systems > 1)
           and not (sub_structure.num_sub_systems == self.num_sub_systems)):

            raise ValueError('The number of subsystems in each sub_structure'
                             'must be the same!')
        else:
            self.num_sub_systems = sub_structure.num_sub_systems

        # add a sub_structure of N repetitions to the structure with
        self.sub_structures.append([sub_structure, N])

    def add_substrate(self, sub_structure):
        """add_substrate

        Add a structure as static substrate to the structure.

        Args:
            sub_structure (Structure): substrate structure.

        """
        if not isinstance(sub_structure, Structure):
            raise ValueError('Class '
                             + type(sub_structure).__name__
                             + ' is no possible substrate. '
                             + 'Only structure class is allowed!')

        self.substrate = sub_structure

    def get_number_of_sub_structures(self):
        """get_number_of_sub_structures

        This methods does not return the number of all layers in the
        structure, see :meth:`.get_number_of_layers`.

        Returns:
            N (int): number of all sub structures.

        """
        N = 0
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], (AmorphousLayer, UnitCell)):
                N = N + 1
            else:
                N = N + self.sub_structures[i][0].get_number_of_sub_structures()
        return N

    def get_number_of_layers(self):
        """get_number_of_layers

        Determines the number of all layers in the structure.

        Returns:
            L (int): number of all layers in the structure.

        """
        L = 0
        # traverse the substructures
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], AmorphousLayer) or \
                    isinstance(self.sub_structures[i][0], UnitCell):
                L = L + self.sub_structures[i][1]
            else:
                # its a structure, so call the method recursively
                L = L + self.sub_structures[i][0].get_number_of_layers() \
                    * self.sub_structures[i][1]

        return L

    def get_number_of_unique_layers(self):
        """get_number_of_unique_layers

        Determines the number of unique layers in the structure.

        Returns:
            N (int): number of unique layers in the structure.

        """
        N = len(self.get_unique_layers()[0])
        return N

    def get_thickness(self, units=True):
        """get_thickness

        Determines the thickness of the structure.

        Args:
            units (boolean, optional): whether units should be returned or not.
                Defaults to True.

        Returns:
            thickness (float, Quantity): the thickness from surface to bottom
            of the structure.

        """
        _, d_end, _ = self.get_distances_of_layers(units)
        return d_end[-1]

    def get_unique_layers(self):
        """get_unique_layers

        The uniqueness is determined by the handle of each layer instance.

        Returns:
            (tuple):
            - *layer_ids (list[str])* - ids of all unique layers instances in
              the structure.
            - *layer_handles (list[AmorphousLayer, UnitCell, Structure])* -
              handles of all unique layers instances in the structure.

        """
        layer_ids = []
        layer_handles = []
        # traverse the sub_structures
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], (AmorphousLayer)) or \
                    isinstance(self.sub_structures[i][0], (UnitCell)):
                # its a AmorphousLayer or UnitCell
                layer_id = self.sub_structures[i][0].id
                if not layer_ids:
                    # the list is empty at the beginning so add
                    # the first layer
                    layer_ids = layer_ids + [layer_id]
                    layer_handles = layer_handles + [self.sub_structures[i][0]]
                else:
                    # the list is not empty so check if the id is
                    # already in the layers id list
                    if layer_id not in layer_ids:
                        # if id not in list, so add it
                        layer_ids = layer_ids + [layer_id]
                        layer_handles = layer_handles + [self.sub_structures[i][0]]
            else:
                # its a sub_structure
                if not layer_ids:
                    # the list is empty at the beginning so call
                    # the method recursively and add the result to the
                    # layers list
                    layer_ids = self.sub_structures[i][0].get_unique_layers()[0]
                    layer_handles = self.sub_structures[i][0].get_unique_layers()[1]
                else:
                    # the list is not empty so check if the ids
                    # from the recursive call are already in the layers id
                    # list.
                    temp1 = self.sub_structures[i][0].get_unique_layers()[0]
                    temp2 = self.sub_structures[i][0].get_unique_layers()[1]
                    for j in range(len(temp1)):
                        # check all ids from recursive call
                        if temp1[j] not in layer_ids:
                            # ids not in list, so add them
                            layer_ids = layer_ids + [temp1[j]]
                            layer_handles = layer_handles + [temp2[j]]

        return layer_ids, layer_handles

    def get_layer_vectors(self, *args):
        """get_layer_vectors

        Returns three lists with the numeric index of all layers
        in a structure given by the get_unique_layers() method and
        additionally vectors with the ids and Handles of the
        corresponding layer instances.
        The list and order of the unique layers can be either handed
        as an input parameter or is requested at the beginning.

        Args:
            layers (Optional[list]): list of unique layers including
               ids and handles

        Returns:
            (tuple):
            - *indices (list[int])* - numeric index of all layers in a
              structure.
            - *layer_ids (list[str])* - ids of all unique layers instances in
              the structure.
            - *layer_handles (list[AmorphousLayer, UnitCell, Structure])* -
              handles of all unique layers instances in the structure.

        """
        indices = []
        layer_ids = []
        layer_handles = []
        # if no layers (unique layers) are given, we have to get them
        if len(args) < 1:
            layers = self.get_unique_layers()
        else:
            layers = args[0]
        # traverse the substructres
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], (AmorphousLayer, UnitCell)):
                # its a AmorphousLayer or UnitCell
                # find the index of the current layer id in the unique
                # layer list
                Index = layers[0].index(self.sub_structures[i][0].id)
                # add the index N times to the indices vector
                indices = np.append(indices, Index*np.ones(self.sub_structures[i][1]))
                # create a list of N layer ids and add them to
                # the ids list
                temp1 = list(itertools.repeat(self.sub_structures[i][0].id,
                                              self.sub_structures[i][1]))
                layer_ids = layer_ids + list(temp1)
                # create a list of N layer handles and add them to
                # the Handles list
                temp2 = list(itertools.repeat(self.sub_structures[i][0],
                                              self.sub_structures[i][1]))
                layer_handles = layer_handles + list(temp2)
            else:
                # its a structure
                # make a recursive call and hand in the same unique
                # layer vector as we used before
                [temp1, temp2, temp3] = self.sub_structures[i][0].get_layer_vectors(layers)
                temp11 = []
                temp22 = []
                temp33 = []
                # concat the temporary arrays N times
                for _ in range(self.sub_structures[i][1]):
                    temp11 = temp11 + list(temp1)
                    temp22 = temp22 + list(temp2)
                    temp33 = temp33 + list(temp3)
                # add the temporary arrays to the outputs
                indices = np.append(indices, temp11)
                layer_ids = layer_ids + list(temp22)
                layer_handles = layer_handles + list(temp33)
        return indices, layer_ids, layer_handles

    def get_all_positions_per_unique_layer(self):
        """get_all_positions_per_unique_layer

        Determines the position indices for each unique layer in the structure.

        Returns:
            pos (dict{ndarray[int]}): position indices for each unique layer in
            the structure.

        """
        layers = self.get_unique_layers()
        indices = self.get_layer_vectors()[0]
        pos = {}  # Dictionary used instead of array
        for i, layer in enumerate(layers[0]):
            pos[layer] = np.flatnonzero(indices == i)
        # Each element accessible through layer id
        return pos

    def get_distances_of_layers(self, units=True):
        """get_distances_of_layers

        Returns a vector of the distance from the surface for each layer
        starting at 0 (dStart) and starting at the end of the first
        layer (dEnd) and from the center of each layer (dMid).

        Args:
            units (boolean, optional): whether units should be returned or not.
                Defaults to True.

        Returns:
            (tuple):
            - *d_start (ndarray[float, Quantity])* - distances from the surface
              of each layer starting at 0.
            - *d_end (ndarray[float, Quantity])* - distances from the bottom
              of each layer.
            - *d_mid (ndarray[float, Quantity])*: distance from the middle of
              each layer.

        """
        thickness = self.get_layer_property_vector('_thickness')
        d_end = np.cumsum(thickness)
        d_start = np.hstack([[0], d_end[0:-1]])
        d_mid = (d_start + thickness/2)
        if units:
            return Q_(d_start, u.m), Q_(d_end, u.m), Q_(d_mid, u.m)
        else:
            return d_start, d_end, d_mid

    def get_distances_of_interfaces(self, units=True):
        """get_distances_of_interfaces

        Calculates the distances of the interafaces of the structure.

        Args:
            units (boolean, optional): whether units should be returned or not.
                Defaults to True.

        Returns:
            res (ndarray[float, Quantity]): distances from the surface of each
            interface of the structure.

        """

        d_start, d_end, _ = self.get_distances_of_layers(False)
        indices = np.r_[1, np.diff(self.get_layer_vectors()[0])]
        res = np.append(d_start[np.nonzero(indices)], d_end[-1])
        if units:
            return Q_(res, u.m)
        else:
            return res

    def interp_distance_at_interfaces(self, N, units=True):
        """ interp_distance_at_interfaces

        Interpolates the distances at the layer interfaces by an odd number
        :math:`N`.

        Args:
            N (int): number of point of interpolation at interface
            units (boolean, optional): whether units should be returned or not.
                Defaults to True.

        Returns:
            (tuple):
            - *dist_interp (ndarray[float, Quantity])* - distance array of the
              middle of each layer interpolated by an odd number :math:`N` at
              the interfaces
            - *original_indicies (ndarray[int])* - indicies of the original
              distances in the interpolated array

        """

        [d_start, d_end, d_mid] = self.get_distances_of_layers(False)
        # these are the distances of the interfaces
        dist_intf = self.get_distances_of_interfaces(False)
        # start with the distances of the centers of the layers
        dist_interp = d_mid

        N = int(N)  # make N an integer
        if N % 2 == 0:
            # odd numbers are required
            N += 1

        # traverse all distances
        for z in dist_intf:
            inda = finderb(z, d_start)  # this is the index of a layer after the interface
            indb = inda-1  # this is the index of a layer before the interface

            # interpolate linearly N new distances at the interface
            if indb == -1:  # this is the surface interface
                dist_interp = np.append(dist_interp, np.linspace(0, d_mid[inda], int(2+(N-1)/2)))
            elif inda >= (len(d_mid)-1):  # this is the bottom interface
                dist_interp = np.append(dist_interp,
                                        np.linspace(d_mid[inda], d_end[-1], int(2+(N-1)/2)))
            else:  # this is a surface inside the structure
                dist_interp = np.append(dist_interp, np.linspace(d_mid[indb], d_mid[inda], 2+N))

        dist_interp = np.unique(np.sort(dist_interp))  # sort and unify the distances
        # these are the indicies of the original distances in the interpolated new array
        original_indicies = finderb(d_mid, dist_interp)
        if units:
            return Q_(dist_interp, u.m), original_indicies
        else:
            return dist_interp, original_indicies

    def get_layer_property_vector(self, property_name):
        """get_layer_property_vector

        Returns a vector for a property of all layers in the
        structure. The property is determined by the propertyName and
        returns a scalar value or a function handle.

        Args:
            property_name (str): name of property to return as array

        Returns:
            prop (ndarray[float, @lambda]): array of a property for all layers
            in the structure.

        """
        # get the Handle to all layers in the Structure
        handles = self.get_layer_vectors()[2]

        if callable(getattr(handles[0], property_name)):
            # it's a function
            prop = np.zeros([self.get_number_of_layers()])
            for i in range(self.get_number_of_layers()):
                prop[i] = getattr(handles[i], property_name)
        elif ((type(getattr(handles[0], property_name)) is list) or
                (type(getattr(handles[0], property_name)) is str) or
                (type(getattr(handles[0], property_name)) is dict)):
            # it's a list of functions or str
            prop = []
            for i in range(self.get_number_of_layers()):
                # Prop = Prop + getattr(Handles[i],types)
                prop.append(getattr(handles[i], property_name))
        elif type(getattr(handles[0], property_name)) is Q_:
            # its a pint quantity
            unit = getattr(handles[0], property_name).units
            prop = np.empty([self.get_number_of_layers()])
            for i in range(self.get_number_of_layers()):
                prop[i] = getattr(handles[i], property_name).magnitude
            prop *= unit
        else:
            # its a number or array
            layers = self.get_unique_layers()
            temp = np.zeros([len(layers[0]), 1])
            set_dtype = float
            for i, layer in enumerate(layers[1]):
                value = getattr(layer, property_name)
                if np.iscomplexobj(value):
                    set_dtype = np.complex128
                try:
                    temp[i] = len(value)
                except TypeError:
                    temp[i] = 1
            max_dim = int(np.max(temp))
            if max_dim > 1:
                prop = np.empty([self.get_number_of_layers(), max_dim], dtype=set_dtype)
            else:
                prop = np.empty([self.get_number_of_layers()], dtype=set_dtype)
            del temp
            # traverse all layers
            for i in range(self.get_number_of_layers()):
                temp = getattr(handles[i], property_name)
                if np.iscomplexobj(temp):
                    prop.dtype = np.complex128
                if max_dim > 1:
                    prop[i, :] = temp
                else:
                    prop[i] = np.asarray(temp).item()

        return prop

    def get_numel_of_layer_property(self, property_name):
        """get_numel_of_layer_property

        Returns the number of elements (numel) of a property within
        the structure.
        For most properties this is 1, but for temperature-dependent
        properties it should return the number of sub-systems.

        The method raises an `IndexError` if the dimension of the property
        is not equal for all layers.

        Args:
            property_name (str): name of property to be checked

        Returns:
            numel (float): number of elements of property.

        """
        numel = np.array([len(prop) for prop in self.get_layer_property_vector(property_name)])
        if np.all(numel == numel[0]):
            numel = numel[0]
        else:
            raise IndexError('Property {:s} has not the same number of elements '
                             '(num_sub_systems) across the whole sample '
                             'structure.'.format(property_name))
        return numel

    def get_layer_handle(self, i):
        """get_layer_handle

        Returns the handle to a layer at a given position index.

        Args:
            i (int): index of the layer to return.

        Returns:
            handle (AmorphousLayer, UnitCell): handle to the layer at position
            `i` in the structure.

        """
        handles = self.get_layer_vectors()[2]
        return handles[i]

    def reverse(self):
        """reverse

        Returns a reversed structure also reversing all nested sub_structure.

        Returns:
            reversed (Structure): reversed structure.

        """
        from copy import deepcopy

        reversed = deepcopy(self)
        # need to handle superstrate and substrate
        return self.reverse_sub_structures(reversed)

    def reverse_sub_structures(self, structure):
        """reverse_sub_structures

        Reverse a `Structure` and recursively call itself if a
        sub_structure is a `Structure` itself.

        Args:
            structure (Structure): structure to be reversed.

        Returns:
            structure (Structure): reversed structure.

        """
        # reverse the list of sub_structures
        structure.sub_structures.reverse()
        for (sub_structure, N) in structure.sub_structures:
            if isinstance(sub_structure, Structure):
                # recursive call
                self.reverse_sub_structures(sub_structure)
            else:
                pass
        return structure



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import u
import numpy as np
import pytest
from pint.testing import assert_allclose as assert_approx
from pint.testing import assert_equal


# tests

# Layer

def test_layer_id(layer):
    assert layer.id == 'layer'


def test_layer_name(layer):
    assert layer.name == 'base layer'


def test_layer_roughness(layer):
    assert_equal(layer.roughness, 0.5*u.angstrom)


def test_layer_deb_wal_fac(layer):
    assert_equal(layer.deb_wal_fac[0](300), 1e-20)


def test_layer_sound_vel(layer):
    assert_equal(layer.sound_vel, 6*u.nm/u.ps)


def test_layer_phonon_damping(layer):
    assert_equal(layer.phonon_damping, 1*u.kg/u.s)


def test_layer_opt_pen_depth(layer):
    # asserat_equal does not work on because of last digit uncertainty
    assert_approx(layer.opt_pen_depth, 10*u.nm)


def test_layer_opt_ref_index(layer):
    assert layer.opt_ref_index == 5-3j


def test_layer_opt_ref_index_per_strain(layer):
    assert layer.opt_ref_index_per_strain == 1-1j


def test_layer_therm_cond(layer):
    assert layer.therm_cond[0](300) == 1


def test_layer_therm_cond_str(layer):
    assert layer.therm_cond_str[0] == '1.0'


def test_layer_lin_therm_exp(layer):
    assert layer.lin_therm_exp[0](300) == 1e-5


def test_layer_lin_therm_exp_str(layer):
    assert layer.lin_therm_exp_str[0] == '1e-05'


def test_layer_int_lin_therm_exp(layer):
    assert layer.int_lin_therm_exp[0](300) == 3e-3


def test_layer_int_lin_therm_exp_str(layer):
    assert layer.int_lin_therm_exp_str[0] == '1.0e-5*T'


def test_layer_heat_capacity(layer):
    assert layer.heat_capacity[0](300) == 10


def test_layer_heat_capacity_str(layer):
    assert layer.heat_capacity_str[0] == '10.0'


def test_layer_int_heat_capacity(layer):
    assert layer.int_heat_capacity[0](300) == 3000.0


def test_layer_int_heat_capacity_str(layer):
    assert layer.int_heat_capacity_str[0] == '10.0*T'


def test_layer_sub_system_coupling(layer):
    assert layer.sub_system_coupling[0](0) == 0


def test_layer_num_subsystems(layer):
    assert layer.num_sub_systems == 1


def test_layer_eff_spin(layer):
    assert layer.eff_spin == 1


def test_layer_curie_temp(layer):
    assert_equal(layer.curie_temp, 100*u.K)


def test_layer_mf_exch_coupling(layer):
    assert_approx(layer.mf_exch_coupling, 2.07e-21*u.m**2*u.kg/u.s**2, rtol=1e-2)


def test_layer_lamda(layer):
    assert layer.lamda == 1


def test_layer_mag_moment(layer):
    # asserat_equal does not work on because of last digit uncertainty
    assert_approx(layer.mag_moment, 1*u.bohr_magneton, rtol=1e-2)


def test_layer_aniso_exponent(layer):
    assert layer.aniso_exponent == 1


def test_layer_anisotropy(layer):
    assert_equal(layer.anisotropy, [1, 2, 3]*u.J/u.m**3)


def test_layer_exch_stiffness(layer):
    assert_equal(layer.exch_stiffness, 1*u.J/u.m)


def test_layer_mag_saturation(layer):
    assert_equal(layer.mag_saturation, 1*u.J/u.T/u.m**3)


@pytest.mark.parametrize('input, expected, expected_str',
                         [(1, 1.0, '1.0'),
                          (1.0, 1.0, '1.0'),
                          (1.0*u.m, 1.0, '1.0'),
                          (1.0*u.nm, 1.0e-9, '1e-09'),
                          ('1', 1.0, '1'),
                          ('1.0', 1.0, '1.0'),
                          ('1.0*T', 300.0, '1.0*T'),
                          ('1.0*T_0', 300.0, '1.0*T_0'),
                          ('lambda T: 1.0*T', 300.0, '1.0*T'),
                          ])
def test_layer_check_input(layer, input, expected, expected_str):
    output, output_str = layer.check_input(input)
    if isinstance(input, str) and '_' in input:
        assert output[0]([300]) == expected
    else:
        assert output[0](300) == expected
    assert output_str[0] == expected_str


@pytest.mark.parametrize('input, expected, expected_str',
                         [([1, 2, 3], [1, 2, 3], ['1.0', '2.0', '3.0']),
                          ([1.0, 2.0, 3.0], [1.0, 2.0, 3.0], ['1.0', '2.0', '3.0']),
                          ([1.0*u.m, 2.0*u.m, 3.0*u.m], [1.0, 2.0, 3.0], ['1.0', '2.0', '3.0']),
                          ([1.0*u.nm, 2.0*u.nm, 2.0*u.nm], [1.0e-9, 2.0e-9, 2.0e-9],
                           ['1e-09', '2e-09', '2e-09']),
                          (['1', '2', '3'], [1.0, 2.0, 3.0], ['1', '2', '3']),
                          (['1.0', '2.0', '3.0'], [1.0, 2.0, 3.0],
                           ['1.0', '2.0', '3.0']),
                          (['1.0*T', '2.0*T', '3.0*T'], [300.0, 600.0, 900.0],
                           ['1.0*T', '2.0*T', '3.0*T']),
                          (['1.0*T_0', '2.0*T_1', '3.0*T_2'], [300.0, 600.0, 900.0],
                           ['1.0*T_0', '2.0*T_1', '3.0*T_2']),
                          (['lambda T: 1.0*T', 'lambda T: 2.0*T', 'lambda T: 3.0*T'],
                           [300.0, 600.0, 900.0], ['1.0*T', '2.0*T', '3.0*T']),
                          ])
def test_layer_check_input_list(layer, input, expected, expected_str):
    output, output_str = layer.check_input(input)
    for i, _input in enumerate(input):
        if isinstance(_input, str) and '_' in _input:
            assert output[i]([300, 300, 300]) == expected[i]
        else:
            assert output[i](300) == pytest.approx(expected[i], rel=1e-2)
        assert output_str[i] == expected_str[i]


def test_layer_get_property_dict(layer):
    assert layer.get_property_dict(types='all') == vars(layer)
    layer.get_property_dict(types='heat')
    layer.get_property_dict(types='phonon')
    layer.get_property_dict(types='xray')
    layer.get_property_dict(types='optical')
    layer.get_property_dict(types='magnetic')


def test_layer_set_ho_spring_constants(layer):
    assert layer.spring_const == 0.
    layer.set_ho_spring_constants([1.0, 2.0, 3.0])
    assert np.allclose(layer.spring_const, [0., 1.0, 2.0, 3.0])


def test_layer_set_opt_pen_depth_from_ref_index(layer):
    layer.set_opt_pen_depth_from_ref_index(800*u.nm)
    assert_approx(layer.opt_pen_depth, 21.22*u.nm, rtol=1e-2)


# AmorphousLayer & UnitCell (need atoms, mass, area to be set)


@pytest.mark.parametrize('fixture_name',
                         ['amorphous_layer', 'unit_cell'])
def test_layer_to_str(request, fixture_name):
    layer = request.getfixturevalue(fixture_name)
    layer.__str__()


@pytest.mark.parametrize('fixture_name, expected',
                         [('amorphous_layer', 1.0*u.nm),
                          ('unit_cell', 5.0*u.angstrom),])
def test_layer_thickness(request, fixture_name, expected):
    layer = request.getfixturevalue(fixture_name)
    assert_approx(layer.thickness, expected, rtol=1e-2)


@pytest.mark.parametrize('fixture_name, expected',
                         [('amorphous_layer', 5e-26*u.kg),
                          ('unit_cell', 2.51e-25*u.kg),])
def test_layer_mass(request, fixture_name, expected):
    layer = request.getfixturevalue(fixture_name)
    assert_approx(layer.mass, expected, rtol=1e-2)


@pytest.mark.parametrize('fixture_name, expected',
                         [('amorphous_layer', 1e-20*u.m**2),
                          ('unit_cell', 2.5e-19*u.m**2),])
def test_layer_area(request, fixture_name, expected):
    layer = request.getfixturevalue(fixture_name)
    assert_approx(layer.area, expected, rtol=1e-2)


@pytest.mark.parametrize('fixture_name, expected',
                         [('amorphous_layer', 1e-29*u.m**3),
                          ('unit_cell', 1.25e-28*u.m**3),])
def test_layer_volume(request, fixture_name, expected):
    layer = request.getfixturevalue(fixture_name)
    assert_approx(layer.volume, expected, rtol=1e-2)


@pytest.mark.parametrize('fixture_name, expected',
                         [('amorphous_layer', 30000000.0*u.kg**0.5/u.m**2),
                          ('unit_cell', 2414871.22*u.kg**0.5/u.m**2),])
def test_layer_get_acoustic_impedance(request, fixture_name, expected):
    layer = request.getfixturevalue(fixture_name)
    assert_approx(layer.get_acoustic_impedance(), expected, rtol=1e-2)


@pytest.mark.parametrize('fixture_name, expected',
                         [('amorphous_layer', 1.8),
                          ('unit_cell', 1.45),])
def test_layer_calc_spring_const(request, fixture_name, expected):
    layer = request.getfixturevalue(fixture_name)
    layer.calc_spring_const()
    assert_approx(layer.spring_const[0], expected, rtol=1e-2)


# AmorphousLayer


def test_amorphous_layer_atom(amorphous_layer, atom_iron):
    assert amorphous_layer.atom == atom_iron


def test_amorphous_layer_magnetization(amorphous_layer):
    assert amorphous_layer.magnetization['amplitude'] == 0.5
    assert_equal(amorphous_layer.magnetization['phi'], 0*u.deg)
    assert_equal(amorphous_layer.magnetization['gamma'], 180*u.deg)


# UnitCell


def test_unit_cell_crystal_axis(unit_cell):
    assert_approx(unit_cell.a_axis, 5.0*u.angstrom, rtol=1e-2)
    assert_approx(unit_cell.b_axis, 5.0*u.angstrom, rtol=1e-2)
    assert_approx(unit_cell.c_axis, 5.0*u.angstrom, rtol=1e-2)


def test_unit_cell_number_atoms(unit_cell):
    assert unit_cell.num_atoms == 3


# test_visualize() fails on certain virtual machine due to TCL error


def test_unit_cell_add_multiple_atoms(unit_cell, atom_oxygen):
    unit_cell.add_multiple_atoms(atom_oxygen, 0.5, 2)


def test_unit_cell_get_atom_ids(unit_cell):
    assert unit_cell.get_atom_ids() == ['Sr', 'O', 'Ti']


def test_unit_cell_get_atom_positions(unit_cell):
    assert np.allclose(unit_cell.get_atom_positions(), [0, 0.5, 0.5, 0.5, 1])
    assert np.allclose(unit_cell.get_atom_positions(0.1), [0, 0.55, 0.55, 0.55, 1.1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = ['Magnetization', 'LLB']

__docformat__ = 'restructuredtext'

from .simulation import Simulation
from ..structures.layers import UnitCell
from .. import u, Q_
from ..helpers import make_hash_md5, finderb
from ..helpers import convert_cartesian_to_polar, convert_polar_to_cartesian
import numpy as np
from scipy.integrate import solve_ivp
from scipy.optimize import fsolve
import scipy.constants as constants
from time import time
from os import path
from tqdm.auto import tqdm


class Magnetization(Simulation):
    """Magnetization

    Base class for all magnetization simulations.

    Args:
        S (Structure): sample to do simulations with.
        force_recalc (boolean): force recalculation of results.

    Keyword Args:
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.

    Attributes:
        S (Structure): sample structure to calculate simulations on.
        force_recalc (boolean): force recalculation of results.
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        ode_options (dict): options for scipy solve_ivp ode solver

    """

    def __init__(self, S, force_recalc, **kwargs):
        super().__init__(S, force_recalc, **kwargs)
        self.ode_options = {
            'method': 'RK45',
            'first_step': None,
            'max_step': np.inf,
            'rtol': 1e-3,
            'atol': 1e-6,
            }

    def __str__(self, output=[]):
        """String representation of this class"""

        class_str = 'Magnetization simulation properties:\n\n'
        class_str += super().__str__(output)
        return class_str

    def get_hash(self, **kwargs):
        """get_hash

        Calculates an unique hash given by the delays as well as the sample
        structure hash for relevant magnetic parameters.
        Optionally, part of the ``strain_map`` and ``temp_map`` are used.

        Args:
            delays (ndarray[float]): delay grid for the simulation.
            **kwargs (ndarray[float], optional): optional delays, strain and
                temperature profile as well as external magnetic field and
                initial magnetization.

        Returns:
            hash (str): unique hash.

        """
        param = []

        if 'strain_map' in kwargs:
            strain_map = kwargs.get('strain_map')
            if np.size(strain_map) > 1e6:
                strain_map = strain_map.flatten()[0:1000000]
            param.append(strain_map)
            kwargs.pop('strain_map')

        if 'temp_map' in kwargs:
            temp_map = kwargs.get('temp_map')
            if np.size(temp_map) > 1e6:
                temp_map = temp_map.flatten()[0:1000000]
            param.append(temp_map)
            kwargs.pop('temp_map')

        for value in kwargs.values():
            param.append(value)

        return self.S.get_hash(types='magnetic') + '_' + make_hash_md5(param)

    def check_initial_magnetization(self, init_mag, distances=[]):
        """check_initial_magnetization

        Check if a given initial magnetization profile is valid. The profile
        must be either a vector `[amplitude, phi, gamma]` describing a global
        initial magnetization or an array of shape `[Nx3]` with N being the
        number of layers or the length of the specific spatial grid. If no
        initial magnetization is given, the initial profile is determined from
        the magnetization of the layers on creation. In addition, a spatial
        grid can be provided.

        Args:
            init_mag (ndarray[float]): initial global or local magnetization
                profile.
            distances (ndarray[float, Quantity], optional): spatial grid of the
                initial magnetization.

        Returns:
            init_mag (ndarray[float]): checked initial magnetization as array on
                the according spatial grid.

        """
        layers = self.S.get_unique_layers()
        for layer in layers[1]:
            if isinstance(layer, UnitCell):
                raise TypeError('UnitCells are not yet supported in Magnetization '
                                'simulations! See issue #129.')

        try:
            distances = distances.to('m').magnitude
        except AttributeError:
            pass

        N = len(distances)
        if N == 0:
            # no spatial grid is provided
            N = self.S.get_number_of_layers()
            [distances, _, _] = self.S.get_distances_of_layers(False)

        if len(init_mag) == 0:
            self.disp_message('No explicit initial magnetization given '
                              '- use magnetization of layers instead.')
            init_mag = np.zeros([N, 3])
            # use finderb search to find the corresponding indices between the
            # internal and external spatial grids
            [d_start, _, _] = self.S.get_distances_of_layers(False)
            idx = finderb(distances, d_start)

            magnetizations = self.S.get_layer_property_vector('_magnetization')
            init_mag[:, 0] = np.array([mag['amplitude'] for mag in magnetizations])[idx]
            init_mag[:, 1] = np.array([mag['phi'] for mag in magnetizations])[idx]
            init_mag[:, 2] = np.array([mag['gamma'] for mag in magnetizations])[idx]
        else:
            if np.size(init_mag) == 3:
                # it is the same initial magnetization for all layers
                init_mag = np.tile(init_mag, (N, 1))
            elif np.shape(init_mag) != (N, 3):
                # init_temp is a vector but has not as many elements as layers
                raise ValueError('The initial magnetization array must have 3 or '
                                 'Nx3 elements, where N is the number of layers '
                                 'in the structure or the length of the spatial '
                                 'grid provided as distances vector!')

            # convert phi and gamma to rad and store only magnitudes
            try:
                init_mag[:, 1] = init_mag[:, 1].to('rad').magnitude
            except AttributeError:
                pass
            try:
                init_mag[:, 2] = init_mag[:, 2].to('rad').magnitude
            except AttributeError:
                pass
        return init_mag

    def get_magnetization_map(self, delays, **kwargs):
        r"""get_magnetization_map

        Returns an absolute ``magnetization_map`` for the sample structure with
        the dimension :math:`M \times N \times 3` with :math:`M` being the
        number of delays and :math:`N` the number of layers in the structure or
        the length of the given spatial grid. Each element of the map contains
        the three magnetization components ``[amplitude, phi, gamma]``.
        The angles ``phi`` and ``gamma`` must be returned in radians as pure
        numpy arrays.
        The ``magnetization_map`` can depend on the ``temp_map`` and
        ``strain_map`` that can be also calculated for the sample structure.
        More over an external magnetic field ``H_ext`` and initial magnetization
        profile ``init_mag`` can be provided.

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            **kwargs (ndarray[float], optional): optional strain and
                temperature profile as well external magnetic field in [T] and
                initial magnetization.

        Returns:
            magnetization_map (ndarray[float]): spatio-temporal absolute
               magnetization profile.

        """
        # create a hash of all simulation parameters
        filename = 'magnetization_map_' \
                   + self.get_hash(delays=delays, **kwargs) \
                   + '.npz'
        full_filename = path.abspath(path.join(self.cache_dir, filename))
        # check if we find some corresponding data in the cache dir
        if path.exists(full_filename) and not self.force_recalc:
            # found something so load it
            tmp = np.load(full_filename)
            magnetization_map = tmp['magnetization_map']
            self.disp_message('_magnetization_map_ loaded from file:\n\t' + filename)
        else:
            t1 = time()
            self.disp_message('Calculating _magnetization_map_ ...')
            # parse the input arguments

            if ('strain_map' in kwargs):
                if not isinstance(kwargs['strain_map'], np.ndarray):
                    raise TypeError('strain_map must be a numpy ndarray!')
            if ('temp_map' in kwargs):
                if not isinstance(kwargs['temp_map'], np.ndarray):
                    raise TypeError('temp_map must be a numpy ndarray!')
            if ('H_ext' in kwargs):
                if not isinstance(kwargs['H_ext'], np.ndarray):
                    raise TypeError('H_ext must be a numpy ndarray!')
                elif kwargs['H_ext'].shape != (3,):
                    raise ValueError('H_ext must be a vector with 3 components '
                                     '(H_x, H_y, H_z)!')
            if ('init_mag' in kwargs):
                if not isinstance(kwargs['init_mag'], np.ndarray):
                    raise TypeError('init_mag must be a numpy ndarray with '
                                    'all in radians without units!')
                elif kwargs['init_mag'].shape != (3,):
                    raise ValueError('init_mag must be a vector with Nx3 '
                                     'with N being the number of layers.')

            magnetization_map = self.calc_magnetization_map(delays, **kwargs)

            self.disp_message('Elapsed time for _magnetization_map_:'
                              ' {:f} s'.format(time()-t1))
            self.save(full_filename, {'magnetization_map': magnetization_map},
                      '_magnetization_map_')
        return magnetization_map

    def calc_magnetization_map(self, delays, **kwargs):
        """calc_magnetization_map

        Calculates an absolute ``magnetization_map`` - see
        :meth:`get_magnetization_map` for details.

        This method is just an interface and should be overwritten for the
        actual simulations.

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            **kwargs (ndarray[float], optional): optional strain and
                temperature profile as well external magnetic field and initial
                magnetization.

        Returns:
            magnetization_map (ndarray[float]): spatio-temporal absolute
                magnetization profile.

        """
        raise NotImplementedError


class LLB(Magnetization):
    """LLB

    Mean-field quantum Landau-Lifshitz-Bloch simulations.

    Please find a detailed review on the Landau-Lifshitz-Bloch equation by
    Unai Atxitia et al. [11]_.

    In collaboration with Theodor Griepe
    (`@Nilodirf <https://github.com/Nilodirf>`_) from the group of
    Unai Atxitia at Instituto de Ciencia de Materiales de Madrid (ICMM-CSIC).

    Args:
        S (Structure): sample to do simulations with.
        force_recalc (boolean): force recalculation of results.

    Keyword Args:
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.

    Attributes:
        S (Structure): sample structure to calculate simulations on.
        force_recalc (boolean): force recalculation of results.
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.

    References:

        .. [11] U. Atxitia, D. Hinzke, and U. Nowak,
           *Fundamentals and Applications of the Landau-Lifshitz-Bloch Equation*,
           `J. Phys. D. Appl. Phys. 50, (2017).
           <https://www.doi.org/10.1088/1361-6463/50/3/033003>`_

    """

    def __init__(self, S, force_recalc, **kwargs):
        super().__init__(S, force_recalc, **kwargs)

    def __str__(self):
        """String representation of this class"""
        class_str = 'Landau-Lifshitz-Bloch Magnetization Dynamics simulation ' \
                    'properties:\n\n'
        class_str += super().__str__()
        return class_str

    def calc_magnetization_map(self, delays, temp_map, H_ext=np.array([0, 0, 0]), init_mag=[]):
        r"""calc_magnetization_map

        Calculates the magnetization map using the mean-field quantum
        Landau-Lifshitz-Bloch equation (LLB) for a given delay range and
        according temperature map:

        .. math::

            \frac{d\mathbf{m}}{dt}=\gamma_e \left(\mathbf{m} \times
              \mathbf{H}_\mathrm{eff} + \frac{\alpha_{\perp}}{m^2}\mathbf{m}
              \times (\mathbf{m} \times \mathbf{H}_\mathrm{eff}) -
              \frac{\alpha_{\parallel}}{m^2}(\mathbf{m} \cdot
              \mathbf{H}_\mathrm{eff}) \cdot \mathbf{m}\right)

        The three terms describe

        #. **precession** at Larmor frequency,
        #. **transversal damping** (conserving the macrospin length), and
        #. **longitudinal damping** (changing macrospin length due to incoherent
           atomistic spin excitations within the layer the macrospin is
           defined on).

        :math:`\alpha_{\parallel}` and :math:`\alpha_{\perp}` are the
        :meth:`longitudinal damping<calc_longitudinal_damping>` and
        :meth:`transverse damping<calc_transverse_damping>` parameters,
        respectively.
        :math:`\gamma_e = -1.761\times10^{11}\,\mathrm{rad\,s^{-1}\,T^{-1}}` is
        the gyromagnetic ratio of an electron.

        The effective magnetic field is the sum of all relevant magnetic
        interactions:

        .. math::

            \mathbf{H}_\mathrm{eff} = \mathbf{H}_\mathrm{ext}
              + \mathbf{H}_\mathrm{A}
              + \mathbf{H}_\mathrm{ex}
              + \mathbf{H}_\mathrm{th}

        where

        * :math:`\mathbf{H}_\mathrm{ext}` is the external magnetic field
        * :math:`\mathbf{H}_\mathrm{A}` is the :meth:`uniaxial anisotropy field
          <calc_uniaxial_anisotropy_field>`
        * :math:`\mathbf{H}_\mathrm{ex}` is the :meth:`exchange field
          <calc_exchange_field>`
        * :math:`\mathbf{H}_\mathrm{th}` is the :meth:`thermal field
          <calc_thermal_field>`

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            temp_map (ndarray[float]): spatio-temporal temperature map.
            H_ext (ndarray[float], optional): external magnetic field
                (H_x, H_y, H_z) [T].

        Returns:
            magnetization_map (ndarray[float]): spatio-temporal absolute
            magnetization profile.

        """
        t1 = time()
        try:
            delays = delays.to('s').magnitude
        except AttributeError:
            pass
        M = len(delays)  # nb of delay steps
        L = self.S.get_number_of_layers()
        K = self.S.num_sub_systems

        temp_map = np.reshape(temp_map, [M, L, K])

        distances, _, _ = self.S.get_distances_of_layers(False)

        init_mag = self.check_initial_magnetization(init_mag, distances)
        # convert initial magnetization from polar to cartesian coordinates
        init_mag = convert_polar_to_cartesian(init_mag)
        # get layer properties
        curie_temps = self.S.get_layer_property_vector('_curie_temp')
        eff_spins = self.S.get_layer_property_vector('eff_spin')
        lambdas = self.S.get_layer_property_vector('lamda')
        mf_exch_couplings = self.S.get_layer_property_vector('mf_exch_coupling')
        mag_moments = self.S.get_layer_property_vector('_mag_moment')
        aniso_exponents = self.S.get_layer_property_vector('aniso_exponent')
        anisotropies = self.S.get_layer_property_vector('_anisotropy')
        mag_saturations = self.S.get_layer_property_vector('_mag_saturation')
        exch_stiffnesses = self.get_directional_exchange_stiffnesses()
        thicknesses = self.S.get_layer_property_vector('_thickness')
        # calculate the mean magnetization maps for each unique layer
        # and all relevant parameters
        mean_mag_map = self.get_mean_field_mag_map(temp_map[:, :, 0])
        # mask for magnetic layers only
        is_magnetic = curie_temps > 0
        N = np.count_nonzero(is_magnetic)

        if self.progress_bar:  # with tqdm progressbar
            pbar = tqdm()
            pbar.set_description('Delay = {:.3f} ps'.format(delays[0]*1e12))
            state = [delays[0], abs(delays[-1]-delays[0])/100]
        else:  # without progressbar
            pbar = None
            state = None
        # solve pdepe with method-of-lines
        sol = solve_ivp(
            LLB.odefunc,
            [delays[0], delays[-1]],
            np.reshape(init_mag[is_magnetic, :], N*3, order='F'),
            args=(delays,
                  N,
                  H_ext,
                  temp_map[:, is_magnetic, 0],  # provide only the electron temperature
                  mean_mag_map[:, is_magnetic],
                  curie_temps[is_magnetic],
                  eff_spins[is_magnetic],
                  lambdas[is_magnetic],
                  mf_exch_couplings[is_magnetic],
                  mag_moments[is_magnetic],
                  aniso_exponents[is_magnetic],
                  anisotropies[is_magnetic],
                  mag_saturations[is_magnetic],
                  exch_stiffnesses[is_magnetic],
                  thicknesses[is_magnetic],
                  pbar, state),
            t_eval=delays,
            **self.ode_options)

        if pbar is not None:  # close tqdm progressbar if used
            pbar.close()
        temp = sol.y.T
        # final magnetization map is zero for all non-magnetic layers
        magnetization_map = np.zeros([M, len(distances), 3])
        # reshape results and set only for magnetic layers
        magnetization_map[:, is_magnetic, :] = np.array(temp).reshape([M, N, 3], order='F')
        # convert to polar coordinates
        magnetization_map = convert_cartesian_to_polar(magnetization_map)
        self.disp_message('Elapsed time for _LLB_: {:f} s'.format(time()-t1))

        return magnetization_map

    def get_mean_field_mag_map(self, temp_map):
        r"""get_mean_field_mag_map

        Returns the mean-field magnetization map see
        :meth:`calc_mean_field_mag_map` for details. The dimension of the map
        are :math:`M \times N` with :math:`M` being the number of delays and
        :math:`N` the number of layers in the structure.

        Args:
            temp_map (ndarray[float]): spatio-temporal electron temperature map.

        Returns:
            mf_mag_map (ndarray[float]): spatio-temporal mean_field
            magnetization map.

        """
        # create a hash of all simulation parameters
        filename = 'mf_magnetization_map_' \
                   + self.get_hash(temp_map=temp_map) \
                   + '.npz'
        full_filename = path.abspath(path.join(self.cache_dir, filename))
        # check if we find some corresponding data in the cache dir
        if path.exists(full_filename) and not self.force_recalc:
            # found something so load it
            tmp = np.load(full_filename)
            mf_mag_map = tmp['mf_mag_map']
            self.disp_message('_mean_field_magnetization_map_ loaded from file:\n\t' + filename)
        else:
            t1 = time()
            self.disp_message('Calculating _mean_field_magnetization_map_ ...')
            # parse the input arguments
            if not isinstance(temp_map, np.ndarray):
                raise TypeError('temp_map must be a numpy ndarray!')

            mf_mag_map = self.calc_mean_field_mag_map(temp_map)

            self.disp_message('Elapsed time for _mean_field_magnetization_map_:'
                              ' {:f} s'.format(time()-t1))
            self.save(full_filename, {'mf_mag_map': mf_mag_map},
                      '_mean_field_magnetization_map_')
        return mf_mag_map

    def calc_mean_field_mag_map(self, temp_map):
        r"""calc_mean_field_mag_map

        Calculate the mean-field magnetization map :math:`m_\mathrm{eq}` by
        solving the self consistent equation

        .. math::

            m_\mathrm{eq}(T) & = B_S(m_\mathrm{eq}, T) \\

        where :math:`B_S` is the Brillouin function.

        Args:
            temp_map (ndarray[float]): spatio-temporal electron temperature map.

        Returns:
            mf_mag_map (ndarray[float]): spatio-temporal mean_field
            magnetization map.

        """
        (M, N) = temp_map.shape  # M - number of delays, N - number of layers

        mf_mag_map = np.zeros_like(temp_map)

        unique_layers = self.S.get_unique_layers()
        relevant_temps = {}
        # iterate over all positions per unique layers
        for i, (k, v) in enumerate(self.S.get_all_positions_per_unique_layer().items()):
            relevant_temps[k] = []
            # unique layer properties
            curie_temp = unique_layers[1][i]._curie_temp
            # mean-field magnetization is only calculated for a non-zero Curie
            # temperature of magnetic layers
            if curie_temp > 0:
                eff_spin = unique_layers[1][i].eff_spin
                mf_exch_coupling = unique_layers[1][i].mf_exch_coupling

                # simple round for down-sampling
                unique_temps = np.unique(np.round(temp_map[:, v].flatten(), decimals=1))
                # only temperatures below T_C are relevant
                unique_temps = unique_temps[unique_temps <= curie_temp]
                #  are normalized by T_C
                reduced_temps = unique_temps/curie_temp
                mf_mags = np.zeros_like(reduced_temps)

                for j, T in enumerate(reduced_temps):
                    if T == 1:
                        mf_mags[j] = 0
                    else:
                        root = fsolve(
                            lambda x: x - LLB.calc_Brillouin(x, T, eff_spin, mf_exch_coupling,
                                                             curie_temp), np.sqrt(1-T))
                        mf_mags[j] = float(root.item())

                relevant_temps[k] = np.stack((unique_temps, mf_mags))

                # for every temperature in temp_map search for best match in
                # relevant_temps and assign according mf_mag into mf_mag_map
                try:
                    idx = finderb(np.round(temp_map[:, v].flatten(), decimals=1),
                                  relevant_temps[k][0, :])
                    mf_mag_map[:, v] = np.reshape(relevant_temps[k][1, idx], (M, len(v)))
                except Exception:
                    raise IndexError('No temperature in _temp_map_ was found that is '
                                     'below the curie temperature for layer {:s}!'.format(
                                         unique_layers[0][i]))
            else:
                # non-magnetic layers with Curie temperature = 0
                mf_mag_map[:, v] = 0

        return mf_mag_map

    def get_directional_exchange_stiffnesses(self):
        r"""get_directional_exchange_stiffnesses

        Returns the directional exchange stiffnesses with size
        :math:`N \times 2`, with :math:`N` being the number of layers, between
        the :math:`(i-1)^\mathrm{th}` and :math:`i^\mathrm{th}` layers as well
        as between the :math:`i^\mathrm{th}` and :math:`(i+1)^\mathrm{th}`
        layers in the structure.
        In case two neighboring layers are identical the center entry
        of the 3-component `exchange_stiffness` is used. In case of an interface
        to the top :math:`(i-1)\rightarrow i` it will be the first entry and
        for an interface to the bottom :math:`i\rightarrow (i+1)` it will be
        the third entry.

        Returns:
            A (ndarray[float]): directional exchange stiffnesses.

        """
        exch_stiffnesses = self.S.get_layer_property_vector('_exch_stiffness')

        indices, _, _ = self.S.get_layer_vectors()

        A = np.zeros([len(indices), 2])
        interfaces = (np.r_[1, np.diff(indices), 1])
        interfaces[interfaces != 0] = -1
        select = (interfaces+1).astype(np.int16)

        A[:, 0] = exch_stiffnesses[np.arange(len(select[0:-1])), select[0:-1]]

        interfaces[interfaces != 0] = 1
        select = (interfaces+1).astype(np.int16)
        A[:, 1] = exch_stiffnesses[np.arange(len(select[1:])), select[1:]]

        return A

    @staticmethod
    def odefunc(t, m,
                delays, N, H_ext, temp_map, mean_mag_map, curie_temps, eff_spins, lambdas,
                mf_exch_couplings, mag_moments, aniso_exponents, anisotropies, mag_saturations,
                exch_stiffnesses, thicknesses, pbar, state):
        """odefunc

        Ordinary differential equation that is solved for 1D LLB.

        Args:
            t (ndarray[float]): internal time steps of the ode solver.
            m (ndarray[float]): internal variable of the ode solver.
            delays (ndarray[float]): delays range of simulation [s].
            N (int): number of spatial grid points.
            H_ext (ndarray[float]): external magnetic field
                (H_x, H_y, H_z) [T].
            temp_map (ndarray[float]): spatio-temporal electron temperature map.
            mean_mag_map (ndarray[float]): spatio-temporal
                mean-field magnetization map.
            curie_temps (ndarray[float]): Curie temperatures of layers.
            eff_spins (ndarray[float]): effective spins of layers.
            lambdas (ndarray[float]): coupling-to-bath parameter of layers.
            mf_exch_couplings (ndarray[float]): mean-field exchange couplings of
                 layers.
            mag_moments (ndarray[float]): atomic magnetic moments of layers.
            aniso_exponents (ndarray[float]): exponent of uniaxial anisotropy of
                layers.
            anisotropies (ndarray[float]): anisotropy vectors of layers.
            mag_saturations (ndarray[float]): saturation magnetization of
                layers.
            exch_stiffnesses (ndarray[float]): exchange stiffness of layers
                towards the upper and lower layer.
            thicknesses (ndarray[float]): thicknesses of layers.
            pbar (tqdm): tqdm progressbar.
            state (list[float]): state variables for progress bar.

        Returns:
            dmdt (ndarray[float]): temporal derivative of internal variable.

        """
        # state is a list containing last updated time t:
        # state = [last_t, dt]
        # I used a list because its values can be carried between function
        # calls throughout the ODE integration
        last_t, dt = state
        try:
            n = int((t - last_t)/dt)
        except ValueError:
            n = 0

        if n >= 1:
            pbar.update(n)
            pbar.set_description('Delay = {:.3f} ps'.format(t*1e12))
            state[0] = t
        elif n < 0:
            state[0] = t

        # initialize arrays
        # reshape input temperature
        m = np.array(m).reshape([N, 3], order='F')

        # nearest delay index for current time t
        idt = finderb(t, delays)[0]
        temps = temp_map[idt, :].flatten()
        # binary masks for layers being under or over its Curie temperature
        under_tc = (temps < curie_temps)
        over_tc = ~under_tc
        # get the current mean-field magnetization
        mf_magnetizations = mean_mag_map[idt, :]

        # actual calculations
        m_squared = np.sum(np.power(m, 2), axis=1)
        gamma_e = -1.761e11

        # external field H_ext is given as input
        # calculate uniaxial anisotropy field
        H_A = LLB.calc_uniaxial_anisotropy_field(m, mf_magnetizations, aniso_exponents,
                                                 anisotropies, mag_saturations)
        # calculate exchange field
        H_ex = LLB.calc_exchange_field(m, exch_stiffnesses, mag_saturations, thicknesses)
        # calculate thermal field
        H_th = LLB.calc_thermal_field(m, m_squared, temps, mf_magnetizations, eff_spins,
                                      curie_temps, mf_exch_couplings, mag_moments, under_tc,
                                      over_tc)

        # calculate the effective field
        H_eff = H_ext + H_A + H_ex + H_th

        # calculate components of LLB
        # precessional term:
        m_rot = np.cross(m, H_eff)

        # damping
        qs = LLB.calc_qs(temps, curie_temps, eff_spins, mf_magnetizations,
                         under_tc)
        # transversal damping
        alpha_trans = LLB.calc_transverse_damping(temps, curie_temps, lambdas,
                                                  qs, mf_magnetizations,
                                                  under_tc, over_tc)
        trans_damping = np.multiply(
            np.divide(alpha_trans, m_squared)[:, np.newaxis],
            np.cross(m, m_rot)
            )
        # longitudinal damping
        alpha_long = LLB.calc_longitudinal_damping(temps, curie_temps,
                                                   eff_spins, lambdas, qs,
                                                   under_tc, over_tc)
        long_damping = np.multiply(
            np.divide(alpha_long, m_squared)[:, np.newaxis],
            np.multiply(np.einsum('ij,ij->i', m, H_eff)[:, np.newaxis], m)
            )

        dmdt = gamma_e * (m_rot + trans_damping - long_damping)

        return np.reshape(dmdt, N*3, order='F')

    @staticmethod
    def calc_uniaxial_anisotropy_field(mag_map, mf_magnetizations, aniso_exponents, anisotropies,
                                       mag_saturations):
        r"""calc_uniaxial_anisotropy_field

        Calculate the uniaxial anisotropy component of the effective field.

        .. math::

            \mathbf{H}_\mathrm{A} = -
            \frac{2}{M_s}
            \left(
                K_x\,m_\mathrm{eq}(T)^{\kappa-2}
                    \begin{bmatrix}0\\m_y\\m_z\end{bmatrix}
                + K_y\,m_\mathrm{eq}(T)^{\kappa-2}
                    \begin{bmatrix}m_x\\0\\m_z\end{bmatrix}
                + K_z\,m_\mathrm{eq}(T)^{\kappa-2}
                    \begin{bmatrix}m_x\\m_y\\0\end{bmatrix}
            \right)

        with :math:`K = (K_x, K_y, K_z)` as the anisotropy and :math:`\kappa` as
        the uniaxial anisotropy exponent.

        Args:
            mag_map (ndarray[float]): spatio-temporal magnetization map
                - possibly for a single delay.
            mf_magnetizations (ndarray[float]): mean-field magnetization of
                layers.
            aniso_exponents (ndarray[float]): exponent of uniaxial
                anisotropy of layers.
            anisotropies (ndarray[float]): anisotropy vectors of layers.
            mag_saturations (ndarray[float]): saturation magnetization of
                layers.

        Returns:
            H_A (ndarray[float]): uniaxial anisotropy field.

        """
        H_A = np.zeros_like(mag_map)

        factor = -2/mag_saturations
        unit_vector = np.array([0, 1, 1])[np.newaxis, :]
        for i in range(3):
            H_A += factor[:, np.newaxis] * anisotropies[:, i, np.newaxis]\
                * np.power(mf_magnetizations,
                           aniso_exponents-2)[:, np.newaxis] \
                * mag_map*np.roll(unit_vector, i, axis=1)

        return H_A

    @staticmethod
    def calc_exchange_field(mag_map, exch_stiffnesses, mag_saturations, thicknesses):
        r"""calc_exchange_field

        Calculate the exchange component of the effective field, which is
        defined as for each :math:`i^\mathrm{th}` layer in the structure as

        .. math::

            H_{\mathrm{ex}, i}=\frac{2}{M_{s,i} \Delta z_i^2}
                \left(A_{i}^{i-1}\left(\mathbf{m}_{i-1}-\mathbf{m}_{i}\right)
                + A_i^{i+1}\left(\mathbf{m}_{i+1}-\mathbf{m}_{i}\right) \right),

        where :math:`\Delta z` is the thickness of the layers or magnetic grains
        and :math:`M_s` is the saturation magnetization. :math:`A_{i}^{i-1}` and
        :math:`A_i^{i+1}` describe the exchange stiffness between the nearest
        neighboring layers provided by
        :meth:`get_directional_exchange_stiffnesses`.

        Args:
            mag_map (ndarray[float]): spatio-temporal magnetization map
                - possibly for a single delay.
            exch_stiffnesses (ndarray[float]): exchange stiffness of layers
                towards the upper and lower layer.
            mag_saturations (ndarray[float]): saturation magnetization of
                layers.

        Returns:
            H_ex (ndarray[float]): exchange field.

        """
        H_ex = np.zeros_like(mag_map)

        m_diff_down = np.concatenate((np.diff(mag_map, axis=0), np.zeros((1, 3))), axis=0)
        m_diff_up = -np.roll(m_diff_down, 1)

        es = np.divide(2, np.multiply(mag_saturations, thicknesses**2))

        H_ex = es[:, np.newaxis]*exch_stiffnesses[:, 0, np.newaxis]*m_diff_up \
            + es[:, np.newaxis]*exch_stiffnesses[:, 1, np.newaxis]*m_diff_down

        return -H_ex

    @staticmethod
    def calc_thermal_field(mag_map, mag_map_squared, temp_map, mf_magnetizations, eff_spins,
                           curie_temps, mf_exch_couplings, mag_moments, under_tc, over_tc):
        r"""calc_thermal_field

        Calculate the thermal component of the effective field.

        .. math::

            \mathbf{H}_\mathrm{th} = \begin{cases}
                \frac{1}{2\chi_{\parallel}}\left(1-\frac{m^2}{m_\mathrm{eq}^2}
                    \right)\mathbf{m} & \mathrm{for}\ T < T_\mathrm{C} \\
                -\frac{1}{\chi_{\parallel}}\left(1+\frac{3}{5}
                    \frac{T_\mathrm{C}}{T-T_\mathrm{C}}m^2\right)\mathbf{m}
                    & \mathrm{for}\ T \geq T_\mathrm{C}
            \end{cases}

        with :math:`\chi_{\parallel}` being the
        :meth:`longitudinal susceptibility<calc_long_susceptibility>`.

        Args:
            mag_map (ndarray[float]): spatio-temporal magnetization map
                - possibly for a single delay.
            mag_map_squared (ndarray[float]): spatio-temporal magnetization map
                squared- possibly for a single delay.
            temp_map (ndarray[float]): spatio-temporal temperature map
                - possibly for a single delay.
            mf_magnetizations (ndarray[float]): mean-field magnetization of
                layers.
            eff_spins (ndarray[float]): effective spin of layers.
            curie_temps (ndarray[float]): Curie temperature of layers.
            mf_exch_couplings (ndarray[float]): mean-field exch. coupling of
                layers.
            mag_moments (ndarray[float]): atomic magnetic moments of layers.
            under_tc (ndarray[boolean]): mask temperatures under the Curie
                temperature.
            over_tc (ndarray[boolean]): mask temperatures over the Curie
                temperature.

        Returns:
            H_th (ndarray[float]): thermal field.

        """
        chi_long = LLB.calc_long_susceptibility(temp_map, mf_magnetizations, curie_temps,
                                                eff_spins, mf_exch_couplings, mag_moments,
                                                under_tc, over_tc)

        H_th = np.zeros_like(temp_map)
        H_th[under_tc] = 1/(2 * chi_long[under_tc]) * (
            1 - mag_map_squared[under_tc]/mf_magnetizations[under_tc]**2
            )
        H_th[over_tc] = -1/chi_long[over_tc] * (
            1 + 3/5 * curie_temps[over_tc]/(temp_map[over_tc]-curie_temps[over_tc])
            ) * mag_map_squared[over_tc]

        return np.multiply(H_th[:, np.newaxis], mag_map)

    @staticmethod
    def calc_Brillouin(mag, temp, eff_spin, mf_exch_coupling, curie_temp):
        r"""calc_Brillouin

        .. math::

            B_S(m, T) = \frac{2 S+1}{2S} \coth{\left(\frac{2S+1}{2S}
            \frac{J \, m}{k_\mathrm{B}\, T}\right)}
            - \frac{1}{2S}\coth{\left(\frac{1}{2S}
            \frac{J \, m}{k_\mathrm{B}\,T}\right)}

        where

        .. math::

            J = 3\frac{S}{S+1}k_\mathrm{B} \, T_\mathrm{C}

        is the mean field exchange coupling constant for effective spin
        :math:`S` and Curie temperature :math:`T_\mathrm{C}`.

        Args:
            mag (ndarray[float]): magnetization of layer.
            temp (ndarray[float]): electron temperature of layer.
            eff_spin (ndarray[float]): effective spin of layer.
            mf_exch_coupling (ndarray[float]): mean-field exch. coupling of
                layers.
            curie_temp (ndarray[float]): Curie temperature of layer.

        Returns:
            brillouin (ndarray[float]): brillouin function.

        """

        eta = mf_exch_coupling.to('m**2*kg/s**2').magnitude * mag / constants.k / temp / curie_temp
        c1 = (2 * eff_spin + 1) / (2 * eff_spin)
        c2 = 1 / (2 * eff_spin)
        brillouin = c1 / np.tanh(c1 * eta) - c2 / np.tanh(c2 * eta)
        return brillouin

    @staticmethod
    def calc_dBrillouin_dx(temp_map, mf_magnetizations, eff_spins, mf_exch_couplings):
        r"""calc_dBrillouin_dx

        Calculate the derivative of the Brillouin function :math:`B_x` at
        :math:`m = m_\mathrm{eq}`:

        .. math::

            B_x = \frac{dB}{dx} = \frac{1}{4S^2\sinh^2(x/2S)}
                -\frac{(2S+1)^2}{4S^2\sinh^2\left(\frac{(2S+1)x}{2S}\right)}

        with :math:`x=\frac{J\,m}{k_\mathrm{B}\,T}`.

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature map
                - possibly for a single delay.
            mf_magnetizations (ndarray[float]): mean-field magnetization of
                layers.
            eff_spins (ndarray[float]): effective spin of layers.
            mf_exch_couplings (ndarray[float]): mean-field exchange couplings of
                layers.

        Returns:
            dBdx (ndarray[float]): derivative of Brillouin function.

        """
        x = np.divide(mf_exch_couplings.to('m**2*kg/s**2').magnitude*mf_magnetizations,
                      constants.k*temp_map)

        two_eff_spins = 2*eff_spins
        dBdx = 1 / (two_eff_spins**2 * np.sinh(x / (two_eff_spins))**2) \
            - (two_eff_spins + 1)**2 / \
            (two_eff_spins**2 * np.sinh(((two_eff_spins + 1) * x) / (two_eff_spins))**2)

        return dBdx

    @staticmethod
    def calc_transverse_damping(temp_map, curie_temps, lambdas, qs,
                                mf_magnetizations, under_tc, over_tc):
        r"""calc_transverse_damping

        Calculate the transverse damping parameter:

        .. math::

            \alpha_{\perp} = \begin{cases}
                \frac{\lambda}{m_\mathrm{eq}(T)}\left(\frac{\tanh(q_s)}{q_s}-
                    \frac{T}{3T_\mathrm{C}}\right)
                    & \mathrm{for}\ T < T_\mathrm{C} \\
                \frac{2 \lambda}{3}\frac{T}{T_\mathrm{C}}
                    & \mathrm{for}\ T \geq T_\mathrm{C}
            \end{cases}

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature map
                - possibly for a single delay.
            curie_temps (ndarray[float]): Curie temperatures of layers.
            lambdas (ndarray[float]): coupling-to-bath parameter of layers.
            qs (ndarray[float]): qs parameter.
            mf_magnetizations (ndarray[float]): mean-field magnetization of
                layers.
            under_tc (ndarray[boolean]): mask temperatures under the Curie
                temperature.
            over_tc (ndarray[boolean]): mask temperatures over the Curie
                temperature.

        Returns:
            alpha_trans (ndarray[float]): transverse damping parameter.

        """
        alpha_trans = np.zeros_like(temp_map)
        alpha_trans[under_tc] = np.multiply(
            np.divide(lambdas[under_tc], mf_magnetizations[under_tc]), (
                np.divide(np.tanh(qs), qs)
                - np.divide(temp_map[under_tc], 3*curie_temps[under_tc])
                )
            )
        alpha_trans[over_tc] = lambdas[over_tc]*2/3*np.divide(
            temp_map[over_tc], curie_temps[over_tc]
            )
        return alpha_trans

    @staticmethod
    def calc_longitudinal_damping(temp_map, curie_temps, eff_spins, lambdas, qs,
                                  under_tc, over_tc):
        r"""calc_transverse_damping

        Calculate the transverse damping parameter:

        .. math::

            \alpha_{\parallel} = \begin{cases}
                \frac{2\lambda}{S+1}
                \frac{1}{\sinh(2q_s)} & \mathrm{for}\ T < T_\mathrm{C} \\
                \frac{2 \lambda}{3}\frac{T}{T_\mathrm{C}}
                    & \mathrm{for}\ T \geq T_\mathrm{C}
            \end{cases}

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature map
                - possibly for a single delay.
            curie_temps (ndarray[float]): Curie temperatures of layers.
            eff_spins (ndarray[float]): effective spins of layers.
            lambdas (ndarray[float]): coupling-to-bath parameter of layers.
            qs (ndarray[float]): qs parameter.
            under_tc (ndarray[boolean]): mask temperatures under the Curie
                temperature.
            over_tc (ndarray[boolean]): mask temperatures over the Curie
                temperature.

        Returns:
            alpha_long (ndarray[float]): transverse damping parameter.

        """
        alpha_long = np.zeros_like(temp_map)
        alpha_long[under_tc] = np.divide(2*np.divide(lambdas[under_tc],
                                                     (eff_spins[under_tc]+1)),
                                         np.sinh(2*qs)
                                         )
        alpha_long[over_tc] = lambdas[over_tc]*2/3*np.divide(
            temp_map[over_tc], curie_temps[over_tc]
            )

        return alpha_long

    @staticmethod
    def calc_qs(temp_map, mf_magnetizations, curie_temps, eff_spins, under_tc):
        r"""calc_qs

        Calculate the :math:`q_s` parameter:

        .. math::

            q_s=\frac{3 T_\mathrm{C} m_\mathrm{eq}(T)}{(2S+1)T}

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature map
                - possibly for a single delay.
            mf_magnetizations (ndarray[float]): mean-field magnetization of
                layers.
            curie_temps (ndarray[float]): Curie temperatures of layers.
            eff_spins (ndarray[float]): effective spins of layers.
            under_tc (ndarray[boolean]): mask temperatures below the Curie
                temperature.

        Returns:
            qs (ndarray[float]): qs parameter.

        """
        return np.divide(
            3*curie_temps[under_tc] * mf_magnetizations[under_tc],
            (2*eff_spins[under_tc] + 1)*temp_map[under_tc]
            )

    @staticmethod
    def calc_long_susceptibility(temp_map, mf_magnetizations, curie_temps, eff_spins,
                                 mf_exch_couplings, mag_moments, under_tc, over_tc):
        r"""calc_long_susceptibility

        Calculate the the longitudinal susceptibility

        .. math::

            \chi_{\parallel} = \begin{cases}
                \frac{\mu_{\rm{B}}\,B_x(m_{eq}, T)}{
                    T\,k_\mathrm{B}-J\,B_x(m_{eq}, T)}
                    & \mathrm{for}\ T < T_\mathrm{C} \\
                \frac{\mu_{\rm{B}}T_\mathrm{C}}{J(T-T_\mathrm{C})}
                    & \mathrm{for}\ T \geq T_\mathrm{C}
            \end{cases}

        with :math:`B_x(m_{eq},T)` being the :meth:`derivative of the Brillouin
        function<calc_dBrillouin_dx>`.

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature map
                - possibly for a single delay.
            mf_magnetizations (ndarray[float]): mean-field magnetization of
                layers.
            curie_temps (ndarray[float]): Curie temperatures of layers.
            eff_spins (ndarray[float]): effective spins of layers.
            mf_exch_couplings (ndarray[float]): mean-field exchange couplings of
                layers.
            mag_moments (ndarray[float]): atomic magnetic moments of layers.
            under_tc (ndarray[boolean]): mask temperatures below the Curie
                temperature.
            over_tc (ndarray[boolean]): mask temperatures over the Curie
                temperature.

        Returns:
            chi_long (ndarray[float]): longitudinal susceptibility.

        """

        dBdx = LLB.calc_dBrillouin_dx(temp_map[under_tc],
                                      mf_magnetizations[under_tc],
                                      eff_spins[under_tc],
                                      mf_exch_couplings[under_tc])

        chi_long = np.zeros_like(temp_map)
        chi_long[under_tc] = np.divide(
            mag_moments[under_tc]*dBdx,
            temp_map[under_tc]*constants.k
            - (mf_exch_couplings.to('m**2*kg/s**2').magnitude)[under_tc]*dBdx
            )
        chi_long[over_tc] = np.divide(
            mag_moments[over_tc]*curie_temps[over_tc],
            (mf_exch_couplings.to('m**2*kg/s**2').magnitude)[over_tc]
            * (temp_map[over_tc]-curie_temps[over_tc])
            )

        return chi_long

    @property
    def distances(self):
        return Q_(self._distances, u.meter).to('nm')

    @distances.setter
    def distances(self, distances):
        self._distances = distances.to_base_units().magnitude



from pint import UnitRegistry
u = UnitRegistry()
u.formatter.default_format = '.4g~P'
Q_ = u.Quantity
from .structures.atoms import Atom, AtomMixed
from .structures.layers import Layer, AmorphousLayer, UnitCell
from .structures.structure import Structure
from .simulations.simulation import Simulation
from .simulations.heat import Heat
from .simulations.phonons import Phonon, PhononNum, PhononAna
from .simulations.magnetization import Magnetization, LLB
from .simulations.xrays import Xray, XrayKin, XrayDyn, XrayDynMag

__all__ = ['Atom', 'AtomMixed', 'Layer', 'AmorphousLayer', 'UnitCell', 'Structure',
           'Simulation', 'Heat', 'Phonon', 'PhononNum', 'PhononAna', 'Magnetization', 'LLB',
           'Xray', 'XrayKin', 'XrayDyn', 'XrayDynDebyeWaller', 'XrayDynMag', 'u', 'Q_']

__version__ = '2.3.0'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import Phonon, PhononNum, PhononAna
from udkm1Dsim import u
import numpy as np
import pytest


# fixtures


@pytest.fixture(scope='module')
def phonon(structure, tmp_path_factory):
    return Phonon(structure, force_recalc=True, cache_dir=tmp_path_factory.mktemp('cache'),
                  save_data=True, disp_messages=True, progress_bar=True,
                  )


@pytest.fixture(scope='module')
def phonon_num(structure, tmp_path_factory):
    return PhononNum(structure, force_recalc=True, cache_dir=tmp_path_factory.mktemp('cache'),
                     save_data=True, disp_messages=True, progress_bar=True,
                     )


@pytest.fixture(scope='module')
def phonon_ana(structure, tmp_path_factory):
    return PhononAna(structure, force_recalc=True, cache_dir=tmp_path_factory.mktemp('cache'),
                     save_data=True, disp_messages=True, progress_bar=True,
                     )


@pytest.fixture(scope='module')
def delays():
    return np.r_[-1:10:0.01]*u.ps


@pytest.fixture(scope='module')
def distances(structure):
    dists, _, _ = structure.get_distances_of_layers()
    return dists


@pytest.fixture(scope='module')
def temp_map(delays, distances):
    return np.zeros((len(delays), len(distances)))


@pytest.fixture(scope='module')
def strain_map(delays, distances):
    return np.zeros((len(delays), len(distances)))


# tests


# Phonon


def test_phonon_str(phonon):
    phonon.__str__()


def test_get_hash(phonon, delays, distances, temp_map):
    phonon.get_hash(delays, temp_map, temp_map)


def test_get_all_strains_per_unique_layer(phonon, strain_map):
    assert np.allclose(phonon.get_all_strains_per_unique_layer(strain_map), [])


def test_get_reduced_strains_per_unique_layer(phonon, strain_map):
    assert np.allclose(phonon.get_reduced_strains_per_unique_layer(strain_map), [])


def test_check_temp_maps(phonon, temp_map, delays):
    phonon.check_temp_maps(temp_map, temp_map, delays)


def test_calc_sticks_from_temp_map(phonon, temp_map):
    phonon.calc_sticks_from_temp_map(temp_map, temp_map)


# PhononNum


def test_phonon_num_str(phonon_num):
    phonon_num.__str__()


def test_phonon_num_get_strain_map(phonon_num, delays, temp_map):
    phonon_num.get_strain_map(delays, temp_map, temp_map)
    phonon_num.force_recalc = False
    phonon_num.get_strain_map(delays, temp_map, temp_map)


# PhononAna


def test_phonon_ana_str(phonon_ana):
    phonon_ana.__str__()


def test_phonon_ana_get_strain_map(phonon_ana, delays, temp_map):
    phonon_ana.get_strain_map(delays, temp_map, temp_map)
    phonon_ana.force_recalc = False
    phonon_ana.get_strain_map(delays, temp_map, temp_map)


def test_phonon_ana_get_energy_per_eigenmode(phonon_ana, delays, temp_map):
    _, A, B = phonon_ana.get_strain_map(delays, temp_map, temp_map)
    phonon_ana.get_energy_per_eigenmode(A, B)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = ['Atom', 'AtomMixed']

__docformat__ = 'restructuredtext'

from .. import u, Q_
import os
import numpy as np
import scipy.constants as constants
import warnings
from tabulate import tabulate


class Atom:
    """Atom

    Smallest structural unit of which larger structures can be build.

    It holds real physical properties of on the atomic level.

    Args:
        symbol (str): symbol of the atom.

    Keyword Args:
        id (str): id of the atom, may differ from symbol and/or name.
        ionicity (int): ionicity of the atom.
        atomic_form_factor_path (str): path to atomic form factor coeffs.
        atomic_form_factor_source (str): either _henke_ or default _chantler_
        magnetic_form_factor_path (str): path to magnetic form factor coeffs.

    Attributes:
        symbol (str): symbol of the element.
        id (str): id of the atom, may differ from symbol and/or name.
        name (str): name of the element (generic).
        atomic_number_z (int): Z atomic number.
        mass_number_a (float): A atomic mass number.
        ionicity (int): ionicity of the atom.
        mass (float): mass of the atom [kg].
        atomic_form_factor_coeff (ndarray[float]): atomic form factor.
           coefficients for energy-dependent atomic form factor.
        cromer_mann_coeff (ndarray[float]): cromer-mann coefficients for
           angular-dependent atomic form factor.
        magnetic_form_factor_coeff (ndarray[float]): magnetic form factor
           coefficients for energy-dependent magnetic form factor.
        mag_amplitude (float): magnetization amplitude -1 .. 1.
        mag_phi (float): phi angle of magnetization [rad].
        mag_gamma (float): gamma angle of magnetization [rad].

    References:

        .. [1] B. L. Henke, E. M. Gullikson & J. C. Davis,
           *X-Ray Interactions: Photoabsorption, Scattering,
           Transmission, and Reflection at E = 50-30,000 eV, Z = 1-92*,
           `Atomic Data and Nuclear Data Tables, 54(2), 181–342, (1993).
           <http://www.doi.org/10.1006/adnd.1993.1013>`_
        .. [2] C.T. Chantler, K. Olsen, R.A. Dragoset, J. Chang, A.R. Kishore,
           S.A. Kotochigova, & D.S. Zucker,
           *Detailed Tabulation of Atomic Form Factors, Photoelectric
           Absorption and Scattering Cross Section, and Mass Attenuation
           Coefficients for Z = 1-92 from E = 1-10 eV to E = 0.4-1.0 MeV*,
           `NIST Standard Reference Database 66.
           <https://dx.doi.org/10.18434/T4HS32>`_
        .. [3] J. Als-Nielson, & D. McMorrow,
           `Elements of Modern X-Ray Physics. New York: John Wiley &
           Sons, Ltd. (2001) <http://www.doi.org/10.1002/9781119998365>`_
        .. [4] D. T. Cromer & J. B. Mann, *X-ray scattering
           factors computed from numerical Hartree–Fock wave functions*,
           `Acta Crystallographica Section A, 24(2), 321–324 (1968).
           <http://www.doi.org/10.1107/S0567739468000550>`_

    """

    def __init__(self, symbol, **kwargs):
        self.symbol = symbol
        self.id = kwargs.get('id', symbol)
        self.ionicity = kwargs.get('ionicity', 0)
        self.mag_amplitude = kwargs.get('mag_amplitude', 0.0)
        self.mag_phi = kwargs.get('mag_phi', 0.0*u.deg)
        self.mag_gamma = kwargs.get('mag_gamma', 0.0*u.deg)

        try:
            filename = os.path.join(os.path.dirname(__file__),
                                    '../parameters/elements.dat')
            symbols = np.genfromtxt(filename, dtype='U2', usecols=(0))
            elements = np.genfromtxt(filename, dtype='U15, i8, f8', usecols=(1, 2, 3))
            [rowidx] = np.where(symbols == self.symbol)
            element = elements[rowidx[0]]
        except Exception as e:
            print('Cannot load element specific data from elements data file!')
            print(e)

        self.name = element[0]
        self.atomic_number_z = element[1]
        self.mass_number_a = element[2]
        self._mass = self.mass_number_a*constants.atomic_mass
        self.mass = self._mass*u.kg
        self.atomic_form_factor_coeff = self.read_atomic_form_factor_coeff(
            filename=kwargs.get('atomic_form_factor_path', ''),
            source=kwargs.get('atomic_form_factor_source', 'chantler'))
        self.magnetic_form_factor_coeff = self.read_magnetic_form_factor_coeff(
            filename=kwargs.get('magnetic_form_factor_path', ''))
        self.cromer_mann_coeff = self.read_cromer_mann_coeff()

    def __str__(self):
        """String representation of this class"""
        output = {'parameter': ['id', 'symbol', 'name', 'atomic number Z', 'mass number A', 'mass',
                                'ionicity', 'Cromer Mann coeff', '', '',
                                'magn. amplitude', 'magn. phi', 'magn. gamma'],
                  'value': [self.id, self.symbol, self.name, self.atomic_number_z,
                            self.mass_number_a, '{:.4g~P}'.format(self.mass.to('kg')),
                            self.ionicity,
                            np.array_str(self.cromer_mann_coeff[0:4]),
                            np.array_str(self.cromer_mann_coeff[4:8]),
                            np.array_str(self.cromer_mann_coeff[8:]),
                            self.mag_amplitude, '{:.4g~P}'.format(self.mag_phi.to('deg')),
                            '{:.4g~P}'.format(self.mag_gamma.to('deg'))]}

        return 'Atom with the following properties\n' + \
               tabulate(output, colalign=('right',), tablefmt='rst', floatfmt=('.2f', '.2f'))

    def read_atomic_form_factor_coeff(self, source='chantler', filename=''):
        """read_atomic_form_factor_coeff

        The coefficients for the atomic form factor :math:`f` in dependence of
        the photon energy :math:`E` is read from a parameter file given by [1]_
        or by [2]_ as default.

        Args:
            source (str, optional): source of atmoic form factors can be either
                _henke_ or _chantler_. Defaults to _chantler_.
            filename (str, optional): full path and filename to the atomic form
                factor coefficients.

        Returns:
            f (ndarray[float]): atomic form factor coefficients.

        """
        if not filename:
            if source not in ['chantler', 'henke']:
                raise ValueError('The source of the atomic form factors must be '
                                 'either chantler or henke!')

            if source == 'chantler':
                sub_path = 'chantler/{:s}.cf'.format(self.symbol.lower())
            elif source == 'henke':
                sub_path = 'henke/{:s}.nff'.format(self.symbol.lower())

            filename = os.path.join(os.path.dirname(__file__),
                                    '../parameters/atomic_form_factors/{:s}'.format(sub_path))
        try:
            f = np.genfromtxt(filename, skip_header=0)
        except FileNotFoundError:
            print('Atomic form factor file {:s} not found!'.format(filename))
            raise

        return f

    @u.wraps(None, (None, 'eV'), strict=False)
    def get_atomic_form_factor(self, energy):
        """get_atomic_form_factor

        The complex atomic form factor for the photon energy :math:`E` [eV] is
        calculated by:

        .. math:: f(E)=f_1 - i f_2

        Convention of Ref. [3]_ (p. 11, footnote) is a negative :math:`f_2`.

        Args:
            energy (ndarray[float]): photon energy [eV].

        Returns:
            f (ndarray[complex]): energy-dependent atomic form factors.

        """
        # interpolate the real and imaginary part in dependence of E
        f1 = np.interp(energy, self.atomic_form_factor_coeff[:, 0],
                       self.atomic_form_factor_coeff[:, 1])
        f2 = np.interp(energy, self.atomic_form_factor_coeff[:, 0],
                       self.atomic_form_factor_coeff[:, 2])

        return f1 - f2*1j

    def read_cromer_mann_coeff(self):
        r"""read_cromer_mann_coeff

        The Cromer-Mann coefficients (Ref. [4]_) are read from a parameter file
        and are returned in the following order:

        .. math:: a_1\; a_2\; a_3\; a_4\; b_1\; b_2\; b_3\; b_4\; c

        Returns:
            cm (ndarray[float]): Cromer-Mann coefficients.

        """
        filename = os.path.join(os.path.dirname(__file__),
                                '../parameters/atomic_form_factors/cromermann.txt')
        try:
            cm = np.genfromtxt(filename, skip_header=1,
                               usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11))
        except FileNotFoundError:
            print('Cromer Mann coefficient file {:s} not found!'.format(filename))
            raise

        return cm[(cm[:, 0] == self.atomic_number_z) & (cm[:, 1] == self.ionicity)][0]

    @u.wraps(None, (None, 'eV', 'm**-1'), strict=False)
    def get_cm_atomic_form_factor(self, energy, qz):
        r"""get_cm_atomic_form_factor

        The atomic form factor :math:`f` is calculated in dependence of the
        photon energy :math:`E` [eV] and the :math:`z`-component of the
        scattering vector :math:`q_z` [Å :math:`^{-1}`] (Ref. [4]_).
        Note that the Cromer-Mann coefficients are fitted for :math:`q_z` in
        [Å :math:`^{-1}`]!

        See Ref. [3]_ (p. 235).

        .. math:: f(q_z,E) = f_{CM}(q_z) + \delta f_1(E) -i f_2(E)

        :math:`f_{CM}(q_z)` is given in Ref. [4]_:

        .. math::

            f_{CM}(q_z) = \sum(a_i \, \exp(-b_i \, (q_z/4\pi)^2))+ c

        :math:`\delta f_1(E)` is the dispersion correction:

        .. math::

            \delta f_1(E) = f_1(E) - \left(\sum^4_i(a_i) + c\right)

        Thus:

        .. math:: f(q_z,E) = \sum(a_i \, \exp(-b_i \, q_z/2\pi))
           + c + f_1(E)-i f_2(E) - \left(\sum(a_i) + c\right)

        .. math:: f(q_z,E) = \sum(a_i \, \exp(-b_i \, q_z/2\pi))
           + f_1(E) -i f_2(E) - \sum(a_i)

        Args:
            energy (ndarray[float]): photon energy [eV].
            qz (ndarray[float]): scattering vector [1/m].

        Returns:
            f (ndarray[complex]): energy- and qz-dependent Cromer-Mann atomic form
            factors.

        """
        # convert from 1/nm to 1/Å and to a real column vector
        qz = np.array(qz*1e10, ndmin=2)
        energy = np.array(energy, ndmin=1)
        if np.size(qz, 0) != len(energy):
            raise TypeError('qz need to have as many rows as energies!')

        f = np.zeros_like(qz, dtype=complex)

        for i, en in enumerate(energy):
            _qz = qz[i, :].reshape(-1, 1)

            f_cm = np.dot(self.cromer_mann_coeff[0:3],
                          np.exp(np.outer(-self.cromer_mann_coeff[4:7],
                                          (_qz/(4*np.pi))**2)))
            f[i, :] = f_cm + self.get_atomic_form_factor(en) -\
                np.sum(self.cromer_mann_coeff[0:3])
        return f

    def read_magnetic_form_factor_coeff(self, filename=''):
        """read_magnetic_form_factor_coeff

        The coefficients for the magnetic form factor :math:`m` in dependence
        of the photon energy :math:`E` is read from a parameter file.

        Args:
            filename (str): optional full path and filename to the magnetic
                form factor coefficients.

        Returns:
            m (ndarray[float]): magnetic form factor coefficients.

        """
        if not filename:
            filename = os.path.join(os.path.dirname(__file__),
                                    '../parameters/magnetic_form_factors/{:s}.mf'.format(
                                            self.symbol))
        try:
            m = np.genfromtxt(filename)
        except FileNotFoundError:
            print('Magnetic form factor file {:s} not found!'.format(filename))
            # return zero array
            m = np.zeros([1, 3])

        return m

    @u.wraps(None, (None, 'eV'), strict=False)
    def get_magnetic_form_factor(self, energy):
        """get_magnetic_form_factor

        The complex magnetic form factor is claculated by:

        .. math:: m(E) = m_1 - i m_2

        for the photon energy :math:`E` [eV].

        Convention of Ref. [3]_ (p. 11, footnote) is a negative :math:`m_2`

        Args:
            energy (ndarray[float]): photon energy [eV].

        Returns:
            m (ndarray[complex]): energy-dependent magnetic form factors.

        """
        # interpolate the real and imaginary part in dependence of E
        m1 = np.interp(energy, self.magnetic_form_factor_coeff[:, 0],
                       self.magnetic_form_factor_coeff[:, 1])
        m2 = np.interp(energy, self.magnetic_form_factor_coeff[:, 0],
                       self.magnetic_form_factor_coeff[:, 2])

        return m1 - m2*1j

    @property
    def mag_phi(self):
        return Q_(self._mag_phi, u.rad).to('deg')

    @mag_phi.setter
    def mag_phi(self, mag_phi):
        self._mag_phi = mag_phi.to_base_units().magnitude

    @property
    def mag_gamma(self):
        return Q_(self._mag_gamma, u.rad).to('deg')

    @mag_gamma.setter
    def mag_gamma(self, mag_gamma):
        self._mag_gamma = mag_gamma.to_base_units().magnitude


class AtomMixed(Atom):
    """AtomMixed

    Representation of mixed atoms in alloys and stochiometric mixtures.

    All properties of the included sub-atoms of class Atom are averaged
    and weighted with their stochiometric ratio.

    Args:
        symbol (str): symbol of the atom.

    Keyword Args:
        id (str): id of the atom, may differ from symbol and/or name.
        name (str): name of the mixed atom, default is symbol.
        atomic_form_factor_path (str): path to atomic form factor coeffs.
        magnetic_form_factor_path (str): path to magnetic form factor coeffs.

    Attributes:
        symbol (str): symbol of the element.
        id (str): id of the atom, may differ from symbol and/or name.
        name (str): name of the mixed atom, default is symbol.
        atomic_number_z (int): Z atomic number.
        mass_number_a (float): A atomic mass number.
        ionicity (int): ionicity of the atom.
        mass (float): mass of the atom [kg].
        atomic_form_factor_coeff (ndarray[float]): atomic form factor.
           coefficients for energy-dependent atomic form factor.
        magnetic_form_factor_coeff (ndarray[float]): magnetic form factor
           coefficients for energy-dependent magnetic form factor.
        mag_amplitude (float): magnetization amplitude -1 .. 1.
        mag_phi (float): phi angle of magnetization [rad].
        mag_gamma (float): gamma angle of magnetization [rad].
        atoms (list[Atoms]): list of Atoms.
        num_atoms (int): number of atoms.

    """

    def __init__(self, symbol, **kwargs):
        self.symbol = symbol
        self.id = kwargs.get('id', symbol)
        self.name = kwargs.get('name', symbol)
        self.mag_amplitude = kwargs.get('mag_amplitude', 0)
        self.mag_phi = kwargs.get('mag_phi', 0*u.deg)
        self.mag_gamma = kwargs.get('mag_gamma', 0*u.deg)
        self.ionicity = 0
        self.atomic_number_z = 0
        self.mass_number_a = 0
        self.mass = 0.0
        self.atoms = []
        self.num_atoms = 0
        self.atomic_form_factor_coeff = self.read_atomic_form_factor_coeff(
            filename=kwargs.get('atomic_form_factor_path', ''))
        self.magnetic_form_factor_coeff = self.read_magnetic_form_factor_coeff(
            filename=kwargs.get('magnetic_form_factor_path', ''))

    def __str__(self):
        """String representation of this class"""

        output = {'parameter': ['id', 'symbol', 'name', 'atomic number Z', 'mass number A', 'mass',
                                'ionicity', 'magn. amplitude', 'magn. phi', 'magn. gamma'],
                  'value': [self.id, self.symbol, self.name, self.atomic_number_z,
                            self.mass_number_a, '{:.4g~P}'.format(self.mass.to('kg')),
                            self.ionicity, self.mag_amplitude,
                            '{:.4g~P}'.format(self.mag_phi.to('deg')),
                            '{:.4g~P}'.format(self.mag_gamma.to('deg'))]}

        output_atom = []
        for i in range(self.num_atoms):
            output_atom.append([self.atoms[i][0].name, '{:.1f} %'.format(self.atoms[i][1]*100)])

        return ('AtomMixed with the following properties\n'
                + tabulate(output, colalign=('right',), tablefmt='rst', floatfmt=('.2f', '.2f'))
                + '\n{:d} Constituents:\n'.format(self.num_atoms)
                + tabulate(output_atom, colalign=('right',), floatfmt=('.2f', '.2f')))

    def add_atom(self, atom, fraction):
        """add_atom

        Add an Atom instance with its stochiometric fraction and recalculate
        averaged properties.

        Args:
            atom (Atom): atom to add.
            fraction (float): fraction of the atom - sum of all fractions must
                be 1.

        """
        if isinstance(atom, Atom):
            self.atoms.append([atom, fraction])
            self.num_atoms = self.num_atoms + 1
            # calculate the mixed atomic properties of the atomMixed instance
            self.atomic_number_z = self.atomic_number_z + fraction * atom.atomic_number_z
            self.mass_number_a = self.mass_number_a + fraction * atom.mass_number_a
            self.mass = self.mass + fraction * atom.mass
            self.ionicity = self.ionicity + fraction * atom.ionicity
        else:
            warnings.warn('Only Atom objects can be added to a MixedAtom!')

    def read_atomic_form_factor_coeff(self, filename=''):
        """read_atomic_form_factor_coeff

        The coefficients for the atomic form factor :math:`f` in dependence of
        the photon energy :math:`E` must be read from an external file given
        by ``filename``.

        Args:
            filename (str, optional): full path and filename to the atomic form
                factor coefficients.

        Returns:
            f (ndarray[float]): atomic form factor coefficients.

        """
        if not filename:
            return None
        try:
            f = np.genfromtxt(filename, skip_header=0)
        except FileNotFoundError:
            print('Atomic form factor file {:s} not found!'.format(filename))
            raise

        return f

    @u.wraps(None, (None, 'eV'), strict=False)
    def get_atomic_form_factor(self, energy):
        """get_atomic_form_factor

        Averaged energy dependent atomic form factor.
        If ``atomic_form_factor_path`` was given on initialization this file
        will be used instead.

        Args:
            energy (ndarray[float]): photon energy [eV].

        Returns:
            f (ndarray[complex]): energy-dependent atomic form factors.

        """
        if self.atomic_form_factor_coeff is None:
            # no external file is given
            # calculate average from added atoms
            f = 0
            for i in range(self.num_atoms):
                f += self.atoms[i][0].get_atomic_form_factor(energy) * self.atoms[i][1]
        else:
            f = super().get_atomic_form_factor(energy)

        return f

    @u.wraps(None, (None, 'eV', 'm**-1'), strict=False)
    def get_cm_atomic_form_factor(self, energy, qz):
        """get_cm_atomic_form_factor

        Averaged energy and qz-dependent atomic form factors.

        Args:
            energy (ndarray[float]): photon energy [eV].
            qz (ndarray[float]): scattering vector [1/m].

        Returns:
            f (ndarray[complex]): energy- and qz-dependent Cromer-Mann atomic
            form factors.

        """
        if self.atomic_form_factor_coeff is None:
            # no external file is given
            # calculate average from added atoms
            f = 0
            for i in range(self.num_atoms):
                f += self.atoms[i][0].get_cm_atomic_form_factor(energy, qz) * self.atoms[i][1]
        else:
            warnings.warn('Cromer-Mann correction cannot be applied to '
                          'atomic form factors from external files. '
                          'Returning uncorrected values instead!')
            f = self.get_atomic_form_factor(energy)

        return f

    def read_magnetic_form_factor_coeff(self, filename=''):
        """read_magnetic_form_factor_coeff

        The coefficients for the magnetic form factor :math:`m` in dependence
        of the photon energy :math:`E` must be read from an external file given
        by ``filename``.

        Args:
            filename (str): optional full path and filename to the magnetic
                form factor coefficients.

        Returns:
            m (ndarray[float]): magnetic form factor coefficients.

        """
        if not filename:
            return None
        try:
            m = np.genfromtxt(filename)
        except FileNotFoundError:
            print('Magnetic form factor file {:s} not found!'.format(filename))
            # return zero array
            m = np.zeros([1, 3])

        return m

    @u.wraps(None, (None, 'eV'), strict=False)
    def get_magnetic_form_factor(self, energy):
        """get_magnetic_form_factor

        Mixed energy dependent magnetic form factors.

        Args:
            energy (ndarray[float]): photon energy [eV].

        Returns:
            f (ndarray[complex]): energy-dependent magnetic form factors.

        """
        if self.magnetic_form_factor_coeff is None:
            # no external file is given
            # calculate average from added atoms
            m = 0
            for i in range(self.num_atoms):
                m += self.atoms[i][0].get_magnetic_form_factor(energy) * self.atoms[i][1]
        else:
            m = super().get_magnetic_form_factor(energy)

        return m

    @property
    def mag_phi(self):
        return Q_(self._mag_phi, u.rad).to('deg')

    @mag_phi.setter
    def mag_phi(self, mag_phi):
        self._mag_phi = mag_phi.to_base_units().magnitude

    @property
    def mag_gamma(self):
        return Q_(self._mag_gamma, u.rad).to('deg')

    @mag_gamma.setter
    def mag_gamma(self, mag_gamma):
        self._mag_gamma = mag_gamma.to_base_units().magnitude
